#
# (C) dGB Beheer B.V.; (LICENSE) http://opendtect.org/OpendTect_license.txt
# AUTHOR   : A. Huck
# DATE     : June 2019
#
# Deep learning apply server
#
#

import sys
import selectors
import json
import io
import numpy as np
import struct

from odpy.common import *


class Message:
    def __init__(self, selector, sock, addr, request):
        self.selector = selector
        self.sock = sock
        self.addr = addr
        self._recv_buffer = b""
        self._send_buffer = b""
        self._payload_len = None
        self._reqid = None
        self._subid = None
        self._jsonheader_len = None
        self.jsonheader = None
        self.response = None
        self.request = request
        self._request_queued = False

    def _set_selector_events_mask(self, mode):
        """Set selector to listen for events: mode is 'r', 'w', or 'rw'."""
        if mode == "r":
            events = selectors.EVENT_READ
        elif mode == "w":
            events = selectors.EVENT_WRITE
        elif mode == "rw":
            events = selectors.EVENT_READ | selectors.EVENT_WRITE
        else:
            raise ValueError(f"Invalid events mask mode {repr(mode)}.")
        self.selector.modify(self.sock, events, data=self)

    def _read(self):
        try:
            # Should be ready to read
            data = self.sock.recv(16777216)
        except BlockingIOError:
            # Resource temporarily unavailable (errno EWOULDBLOCK)
            pass
        else:
            if data:
                self._recv_buffer += data
            else:
                raise RuntimeError("Peer closed.")

    def _write(self):
        if self._send_buffer:
            try:
                # Should be ready to write
                sent = self.sock.send(self._send_buffer)
            except BlockingIOError:
                # Resource temporarily unavailable (errno EWOULDBLOCK)
                pass
            else:
                self._send_buffer = self._send_buffer[sent:]

    def _json_encode(self, obj, encoding):
        json_hdr = json.dumps(obj, ensure_ascii=False).encode(encoding)
        return struct.pack('=i',len(json_hdr)) + json_hdr

    def _json_decode(self, json_bytes, encoding):
        json_hdr = struct.unpack('=i',json_bytes[:4])[0]
        tiow = io.TextIOWrapper(
            io.BytesIO(json_bytes[4:4+json_hdr]), encoding=encoding, newline=""
        )
        obj = json.load(tiow)
        tiow.close()
        return (obj,json_bytes[4+json_hdr:])

    def _array_encode(self, objs):
        ret = bytes()
        shapes = list()
        for obj in objs:
          ret += obj.tobytes()
          shapes.append( obj.shape )
        return (ret,shapes)

    def _array_decode(self, arrptr, shapes, dtypes):
        ret = list()
        offset = 0
        for shape,dtype in zip(shapes,dtypes):
          nrsamples = np.prod(shape,dtype=np.int64)
          arr = np.frombuffer(arrptr,dtype,count=nrsamples,offset=offset)
          arr = arr.reshape( shape )
          offset += arr.nbytes
          ret.append( arr )
        return {
          'result': 'arrays',
          'data': ret
        }

    def _create_message(
        self, *, content_bytes, content_type, content_encoding, arrsize
    ):
        jsonheader = {
            'byteorder': sys.byteorder,
            'content-type': content_type,
            'content-encoding': content_encoding,
            'content-length': len(content_bytes),
        }
        if arrsize != None:
          jsonheader.update({ 'array-shape': arrsize })
        jsonheader_bytes = self._json_encode(jsonheader, 'utf-8')
        od_hdr =   struct.pack('=i',len(jsonheader_bytes)+len(content_bytes)) \
                 + struct.pack('=i',1) \
                 + struct.pack('=h',-1)
        message = od_hdr + jsonheader_bytes + content_bytes
        return message

    def _process_response_json_content(self):
        content = self.response
        result = content.get('result')

    def _process_response_array_content(self):
        content = self.response
        result = content.get('result')
#        for arr in content['data']:
#          log_msg( arr.shape, arr.dtype.name )

    def _process_response_binary_content(self):
        content = self.response

    def process_events(self, mask):
        if mask & selectors.EVENT_READ:
            self.read()
        if mask & selectors.EVENT_WRITE:
            self.write()

    def read(self):
        self._read()

        if self._jsonheader_len is None:
            self.process_protoheader()

        if self._jsonheader_len is not None:
            if self.jsonheader is None:
                self.process_jsonheader()

        if self.jsonheader:
            if self.response is None:
                self.process_response()

    def write(self):
        if not self._request_queued:
            self.queue_request()

        self._write()

        if self._request_queued:
            if not self._send_buffer:
                # Set selector to listen for read events, we're done writing.
                self._set_selector_events_mask("r")

    def close(self):
        try:
            self.selector.unregister(self.sock)
        except Exception as e:
            log_msg(
                f"error: selector.unregister() exception for",
                f"{self.addr}: {repr(e)}",
            )

        try:
            self.sock.close()
        except OSError as e:
            log_msg(
                f"error: socket.close() exception for",
                f"{self.addr}: {repr(e)}",
            )
        finally:
            # Delete reference to socket object for garbage collection
            self.sock = None

    def queue_request(self):
        content = self.request['content']
        content_type = self.request['type']
        content_encoding = self.request['encoding']
        if content_type == 'text/json':
            req = {
                'content_bytes': self._json_encode(content, content_encoding),
                'content_type': content_type,
                'content_encoding': content_encoding,
                'arrsize': None,
            }
        elif content_type == 'binary/array':
            (arrsptr,shapes) = self._array_encode(content)
            req = {
              'content_bytes': arrsptr,
              'content_type': content_type,
              'content_encoding': content_encoding,
              'arrsize': shapes,
            }
        else:
            req = {
                'content_bytes': content,
                'content_type': content_type,
                'content_encoding': content_encoding,
                'arrsize': None,
            }
        message = self._create_message(**req)
        self._send_buffer += message
        self._request_queued = True

    def process_protoheader(self):
        hdrlen = 10
        if len(self._recv_buffer) >= hdrlen:
          self._payload_len = struct.unpack('=i',self._recv_buffer[0:4])[0]
          self._reqid = struct.unpack('=i',self._recv_buffer[4:8])[0]
          self._subid = struct.unpack('=h',self._recv_buffer[8:hdrlen])[0]
          self._jsonheader_len = struct.unpack('=i',self._recv_buffer[hdrlen:14])[0]
          self._recv_buffer = self._recv_buffer[hdrlen:]

    def process_jsonheader(self):
        hdrlen = self._jsonheader_len
        if len(self._recv_buffer) >= hdrlen:
            (self.jsonheader,self._recv_buffer) = self._json_decode(
                self._recv_buffer, "utf-8"
            )
            for reqhdr in (
                "byteorder",
                "content-length",
                "content-type",
                "content-encoding",
            ):
                if reqhdr not in self.jsonheader:
                    raise ValueError(f'Missing required header "{reqhdr}".')

    def process_response(self):
        content_len = self.jsonheader["content-length"]
        if not len(self._recv_buffer) >= content_len:
            return
        data = self._recv_buffer[:content_len]
        self._recv_buffer = self._recv_buffer[content_len:]
        if self.jsonheader["content-type"] == "text/json":
            encoding = self.jsonheader["content-encoding"]
            (self.response,self._recv_buffer) = self._json_decode(data, encoding)
            self._process_response_json_content()
        elif self.jsonheader["content-type"] == 'binary/array':
            shapes = self.jsonheader['array-shape']
            dtypes = self.jsonheader['content-encoding']
            self.response = self._array_decode(data,shapes,dtypes)
            self._process_response_array_content()
        else:
            # Binary or unknown content-type
            self.response = data
            log_msg(
                f'received {self.jsonheader["content-type"]} response from',
                self.addr,
            )
            self._process_response_binary_content()
        # Close when response has been processed
        self.close()



#__________________________________________________________________________
#
# (C) dGB Beheer B.V.; (LICENSE) http://opendtect.org/OpendTect_license.txt
# Author:        A. Huck
# Date:          Apr 2019
#
# _________________________________________________________________________

from enum import Enum
from functools import partial

from bokeh.core import enums
from bokeh.layouts import row, column
from bokeh.models import Spacer
from bokeh.models.widgets import Button
from bokeh.plotting import curdoc

but_width = 80
but_height = 32
but_spacer = 5
go_lbl = '▶ Run'
stop_lbl = '◼ Abort'
pause_lbl = '❚❚ Pause'
resume_lbl = '► Resume'
timerkey = 'timerobj'

RunState = Enum( 'RunState', 'Ready Running Pause', module=__name__ )

def getButton(nm,type=enums.ButtonType.default,callback_fn=None):
  ret = Button(label=nm,button_type=type,width=but_width,height=but_height)
  if callback_fn != None:
    ret.on_click(partial(callback_fn,cb=ret))
  return ret

def getRunStopButton(callback_fn=None):
  return getButton(go_lbl,type=enums.ButtonType.success,callback_fn=callback_fn)

def getPauseResumeButton(callback_fn=None):
  return getButton(pause_lbl,type=enums.ButtonType.primary,callback_fn=callback_fn)

def getRunButtonsBar(runact,abortact,pauseact,resumeact,timercb):
  runstopbut = getRunStopButton()
  pauseresumebut = getPauseResumeButton()
  pauseresumebut.visible = False
  buttonsgrp = row(pauseresumebut,Spacer(width=but_spacer),runstopbut,width_policy='min')
  buttonsfld = row(Spacer(width=but_spacer),buttonsgrp, sizing_mode='stretch_width')
  ret = {
    'run': runstopbut,
    'pause': pauseresumebut,
    'state': RunState.Ready,
    timerkey: None
  }
  runstopbut.on_click(partial(startStopCB,cb=ret,run_fn=runact,abort_fn=abortact,
                              timer_fn=timercb) )
  pauseresumebut.on_click(partial(pauseResumeCB,cb=ret,pause_fn=pauseact,resume_fn=resumeact))
  return buttonsfld

def startStopCB( cb, run_fn, abort_fn, timer_fn, repeat=2000 ):
  if isReady( cb ):
    canrun = run_fn( cb[timerkey] )
    if not canrun:
      return
    setRunning( cb )
    cb.update({
      'cb': curdoc().add_periodic_callback(partial(timerCB,cb=cb,timer_fn=timer_fn),repeat)
    })
  else:
    setReady( cb )
    cb[timerkey] = abort_fn( cb[timerkey] )

def pauseResumeCB( cb, pause_fn, resume_fn ):
  if isRunning( cb ):
    setPaused( cb )
    cb[timerkey] = pause_fn( cb[timerkey] )
  else:
    setResumed( cb )
    cb[timerkey] = resume_fn( cb[timerkey] )

def timerCB( cb, timer_fn ):
  (docontinue,cb) = timer_fn( cb )
  if not docontinue:
    setReady( cb )

def isReady( runbutbar ):
  return runbutbar['state'] == RunState.Ready

def isRunning( runbutbar ):
  return runbutbar['state'] == RunState.Running

def setRunning( runbutbar ):
  runbutbar['state'] = RunState.Running
  runbutbar['run'].label = stop_lbl
  runbutbar['run'].button_type = enums.ButtonType.danger
  runbutbar['pause'].visible = True

def setReady( runbutbar ):
  runbutbar['state'] = RunState.Ready
  runbutbar['run'].label = go_lbl
  runbutbar['run'].button_type = enums.ButtonType.success
  runbutbar['pause'].visible = False
  if 'cb' in runbutbar:
    cb = runbutbar.pop( 'cb' )
    curdoc().remove_periodic_callback( cb )

def setPaused( runbutbar ):
  runbutbar['state'] = RunState.Pause
  runbutbar['pause'].label = resume_lbl

def setResumed( runbutbar ):
  runbutbar['state'] = RunState.Running
  runbutbar['pause'].label = pause_lbl

def setTabFromButton( cb, panelnm, tabnm ):
  setActiveTab( panelnm, tabnm )

def setActiveTab( tabspanelwidget, tabnm ):
  tabs = tabspanelwidget.tabs
  for i in range( len(tabs) ):
    if tabs[i].title == tabnm:
      tabspanelwidget.active = i
      return

def integerListContains( listobj, index ):
  for itm in listobj:
    if itm == index:
      return True
  return False

def getGroup( grpnm, options, groups, keywd ):
  for opt,group in zip(options,groups):
    if opt == grpnm:
      if keywd in group:
        return group[keywd]
      return group
  return None

def getAllUiFlds( objects ):
  ret = list()
  if 'uiobjects' in objects:
    for uiobj in objects['uiobjects']:
      ret.append( objects['uiobjects'][uiobj] )
  else:
    try:
      for obj in objects:
        ret.extend( getAllUiFlds(obj) )
    except TypeError:
      pass
  return ret



#__________________________________________________________________________
#
# (C) dGB Beheer B.V.; (LICENSE) http://opendtect.org/OpendTect_license.txt
# Author:        A. Huck
# Date:          Nov 2018
#
# _________________________________________________________________________
# various tools machine learning using Keras platform
#

import os
import re
import json
from datetime import datetime
import numpy as np
import math
from pathlib import Path

from odpy.common import log_msg, redirect_stdout, restore_stdout
import dgbpy.keystr as dgbkeys
import dgbpy.hdf5 as dgbhdf5
import dgbpy.keras_classes as kc

os.environ['TF_FORCE_GPU_ALLOW_GROWTH'] = 'true'
withtensorboard = True
if 'KERAS_WITH_TENSORBOARD' in os.environ:
  withtensorboard = not ( os.environ['KERAS_WITH_TENSORBOARD'] == False or \
                          os.environ['KERAS_WITH_TENSORBOARD'] == 'No' )
withaugmentation = False
if 'KERAS_WITH_AUGMENTATION' in os.environ:
  withaugmentation = not ( os.environ['KERAS_WITH_AUGMENTATION'] == False or \
                           os.environ['KERAS_WITH_AUGMENTATION'] == 'No' )


platform = (dgbkeys.kerasplfnm,'Keras (tensorflow)')

cudacores = [ '1', '2', '4', '8', '16', '32', '48', '64', '96', '128', '144', '192', '256', \
              '288',  '384',  '448',  '480',  '512',  '576',  '640',  '768', \
              '896',  '960',  '1024', '1152', '1280', '1344', '1408', '1536', \
              '1664', '1792', '1920', '2048', '2176', '2304', '2432', '2496', \
              '2560', '2688', '2816', '2880', '2944', '3072', '3584', '3840', \
              '4352', '4608', '4992', '5120' ]

def getMLPlatform():
  return platform[0]

def getUIMLPlatform():
  return platform[1]

prefercpustr = 'prefercpu'
defbatchstr = 'defaultbatchsz'

keras_dict = {
  dgbkeys.decimkeystr: False,
  'nbchunk': 10,
  'epoch': 15,
  'batch': 32,
  'patience': 5,
  'learnrate': 1e-4,
  'epochdrop': 5,
  'type': None,
  'prefercpu': None
}

def can_use_gpu():
  from tensorflow import config as tfconfig
  return len(tfconfig.list_physical_devices('GPU')) > 0

def get_cpu_preference():
  from tensorflow import config as tfconfig
  return len(tfconfig.list_physical_devices('GPU')) < 1

def get_keras_infos():
  ret = {
    'haskerasgpu': can_use_gpu(),
    prefercpustr: get_cpu_preference(),
    'batchsizes': cudacores,
    defbatchstr: keras_dict['batch']
   }
  return json.dumps( ret )

def set_compute_device( prefercpu=get_cpu_preference() ):
  if not prefercpu:
      return
  from tensorflow import config as tfconfig
  cpudevs = tfconfig.list_physical_devices('CPU')
  tfconfig.set_visible_devices( cpudevs )

def getParams( dodec=keras_dict[dgbkeys.decimkeystr], nbchunk=keras_dict['nbchunk'],
               epochs=keras_dict['epoch'],
               batch=keras_dict['batch'], patience=keras_dict['patience'],
               learnrate=keras_dict['learnrate'],epochdrop=keras_dict['epochdrop'],
               nntype=keras_dict['type'],prefercpu=keras_dict['prefercpu']):
  ret = {
    dgbkeys.decimkeystr: dodec,
    'nbchunk': nbchunk,
    'epoch': epochs,
    'batch': batch,
    'patience': patience,
    'learnrate': learnrate,
    'epochdrop': epochdrop,
    'type': nntype
  }
  if prefercpu == None:
    prefercpu = get_cpu_preference()
  ret.update({'prefercpu': prefercpu})
  if not dodec:
    ret['nbchunk'] = 1
  return ret

# Function that takes the epoch as input and returns the desired learning rate
# input_int: the epoch that is currently being entered
def adaptive_schedule(initial_lrate=keras_dict['learnrate'],
                      epochs_drop=keras_dict['epochdrop']):
  from keras.callbacks import LearningRateScheduler
  def adaptive_lr(input_int):
    drop = 0.5
    return initial_lrate * math.pow(drop,
                                    math.floor((1+input_int)/epochs_drop))
    # return the learning rate (quite arbitrarily decaying)
  return LearningRateScheduler(adaptive_lr)


def get_data_format( model ):
  layers = model.layers
  for i in range(len(layers)):
    laycfg = layers[i].get_config()
    if 'data_format' in laycfg:
      return laycfg['data_format']
  return None

def getCubeletShape( model ):
  data_format = get_data_format( model )
  if data_format == 'channels_first':
    cubeszs = model.input_shape[2:]
  elif data_format == 'channels_last':
    cubeszs = model.input_shape[1:-1]
  return cubeszs

def rm_tree(pth):
    pth = Path(pth)
    for child in pth.glob('*'):
        if child.is_file():
            child.unlink()
        else:
            try:
              rm_tree(child)
            except OSError:
              pass
    try:
      pth.rmdir()
    except OSError:
      pass

def getLogDir( examplenm, basedir, clearlogs, args ):
  if not withtensorboard or basedir == None or not Path(basedir).exists():
    return None
  logdir = Path(basedir) / Path(examplenm).stem
  if logdir.exists():
      if clearlogs:
         for child in logdir.glob('*'):
            rm_tree(child)
  else:
      try:
         logdir.mkdir()
      except:
         return None

  if dgbkeys.surveydictstr in args:
    jobnm = args[dgbkeys.surveydictstr][0] + '_run'
  else:
    jobnm = 'run'

  nrsavedruns = 0
  with os.scandir(logdir) as it:
    for entry in logdir.iterdir():
      if entry.name.startswith(jobnm) and entry.is_dir():
        nrsavedruns += 1
  logdir = logdir / Path(jobnm+str(nrsavedruns+1)+'_'+'m'.join( datetime.now().isoformat().split(':')[:-1] ))
  return logdir

def get_model_shape( shape, nrattribs, attribfirst=True ):
  ret = ()
  if attribfirst:
    ret += (nrattribs,)
  if isinstance( shape, int ):
    ret += (shape,)
    if not attribfirst:
      ret += (nrattribs,)
    return ret
  else:
    for i in shape:
      if i > 1:
        ret += (i,)
  if attribfirst:
    if len(ret) == 1:
      ret += (1,)
  else:
    if len(ret) == 0:
      ret += (1,)
  if not attribfirst:
    ret += (nrattribs,)
  return ret

def getModelDims( model_shape, data_format ):
  if data_format == 'channels_first':
    ret = model_shape[1:]
  else:
    ret = model_shape[:-1]
  if len(ret) == 1 and ret[0] == 1:
    return 0
  return len(ret)

def getModelsByType( learntype, classification, ndim ):
    predtype = kc.DataPredType.Continuous
    outtype = kc.OutputType.Pixel
    dimtype = kc.DimType(ndim)
    if dgbhdf5.isImg2Img(learntype):
        outtype = kc.OutputType.Image
    if classification or dgbhdf5.isSeisClass( learntype ):
            predtype = kc.DataPredType.Classification
    return kc.UserModel.getNamesByType(pred_type=predtype, out_type=outtype, dim_type=dimtype)

def getDefaultModel(setup,type=keras_dict['type'],
                     learnrate=keras_dict['learnrate'],
                     data_format='channels_first'):
  isclassification = setup[dgbhdf5.classdictstr]
  if isclassification:
    nroutputs = len(setup[dgbkeys.classesdictstr])
  else:
    nroutputs = dgbhdf5.getNrOutputs( setup )

  nrattribs = dgbhdf5.getNrAttribs(setup)
  model_shape = get_model_shape( setup[dgbkeys.inpshapedictstr], nrattribs,
                                 attribfirst=data_format=='channels_first' )

  if kc.UserModel.findName(type):
    return kc.UserModel.findName(type).model(model_shape, nroutputs,
                                             learnrate,
                                             data_format=data_format)
  return None


def train(model,training,params=keras_dict,trainfile=None,logdir=None,
          withaugmentation=withaugmentation,tempnm=None):
  redirect_stdout()
  import keras
  from keras.callbacks import EarlyStopping
  from dgbpy.keras_classes import TrainingSequence
  restore_stdout()
  
  infos = training[dgbkeys.infodictstr]
  classification = infos[dgbkeys.classdictstr]
  if classification:
    monitor = 'accuracy'
  else:
    monitor = 'loss'
  early_stopping = EarlyStopping(monitor=monitor, patience=params['patience'])
  LR_sched = adaptive_schedule(params['learnrate'],params['epochdrop'])
  callbacks = [early_stopping,LR_sched]
  batchsize = params['batch']
  if logdir != None:
    from keras.callbacks import TensorBoard
    tensor_board = TensorBoard(log_dir=logdir, \
                               batch_size=batchsize,\
                         write_graph=True, write_grads=False, write_images=True)
    callbacks.append( tensor_board )
  train_datagen = TrainingSequence( training, False, model, exfilenm=trainfile, batch_size=batchsize, with_augmentation=withaugmentation, tempnm=tempnm )
  validate_datagen = TrainingSequence( training, True, model, exfilenm=trainfile, batch_size=batchsize, with_augmentation=withaugmentation )
  nbchunks = len( infos[dgbkeys.trainseldicstr] )
  for ichunk in range(nbchunks):
    log_msg('Starting training iteration',str(ichunk+1)+'/'+str(nbchunks))
    if not train_datagen.set_chunk(ichunk) or not validate_datagen.set_chunk(ichunk):
      continue

    if batchsize == 1:
      log_msg( 'Training on', len(train_datagen), 'samples' )
      log_msg( 'Validate on', len(validate_datagen), 'samples' )
    else:
      log_msg( 'Training on', len(train_datagen), 'batches of', batchsize, 'samples' )
      log_msg( 'Validate on', len(validate_datagen), 'batches of', batchsize, 'samples' )

    redirect_stdout()
    model.fit(x=train_datagen,epochs=params['epoch'],\
              validation_data=validate_datagen,callbacks=callbacks)
    restore_stdout()

  try:
    keras.utils.print_summary( model, print_fn=log_msg )
  except:
    pass
  infos = updateModelShape( infos, model, True )
  infos = updateModelShape( infos, model, False )

  return model

def updateModelShape( infos, model, forinput ):
  if forinput:
    shapekey = dgbkeys.inpshapedictstr
    modelshape = model.input_shape
  else:
    shapekey = dgbkeys.outshapedictstr
    modelshape = model.output_shape
    
  exshape = infos[shapekey]
  if get_data_format(model) == 'channels_first':
    modelshape = modelshape[2:]
  else:
    modelshape = modelshape[1:-1]

  if isinstance(exshape,int):
    return infos

  if len(exshape) == len(modelshape) and \
     np.prod(exshape,dtype=np.int64) == np.prod(modelshape,dtype=np.int64):
    return infos

  ret = ()
  i = 0
  for exdim in exshape:
    if exdim < 2:
      ret += (exdim,)
    else:
      ret += (modelshape[i],)
      i += 1
      
  infos[shapekey] = ret
  return infos

def save( model, outfnm ):
  try:
    model.save( outfnm, save_format='h5' )
  except Exception:
    model.save( outfnm )

def load( modelfnm, fortrain ):
  redirect_stdout()
  from tensorflow.keras.models import load_model
  try:
    ret = load_model( modelfnm, compile=fortrain )
  except ValueError:
    configfile = os.path.splitext( modelfnm )[0] + '.json'
    if not os.path.isfile(configfile):
      return None
    import json
    with open(configfile,'r') as f:
      model_json = json.load(f)
    from keras.models import model_from_json
    try:
      ret = model_from_json(model_json)
    except TypeError:
      model_json_str = json.dumps( model_json )
      ret = model_from_json( model_json_str )
    ret.load_weights(modelfnm)
      
  restore_stdout()
  return ret

def transfer( model ):
  from keras.layers import (Conv1D,Conv2D,Conv3D,Dense)
  layers = model.layers
  for layer in layers:
    layer.trainable = False

  for ilay in range(len(layers)):
    layers[ilay].trainable = True
    laytype = type( layers[ilay] )
    if laytype == Conv3D or laytype == Conv2D or laytype == Conv1D:
       break

  for ilay in range(len(layers)-1,0,-1):
    layers[ilay].trainable = True
    laytype = type( layers[ilay] )
    if laytype == Conv3D or laytype == Conv2D or laytype == Conv1D or laytype == Dense:
      break

  return model

def apply( model, samples, isclassification, withpred, withprobs, \
           withconfidence, doprobabilities, scaler=None, batch_size=None ):
  if batch_size == None:
    batch_size = keras_dict['batch']
  redirect_stdout()
  import keras
  restore_stdout()
  ret = {}
           
  inp_shape = samples.shape
  data_format = 'channels_first'
  samples = adaptToModel( model, samples, sample_data_format=data_format )
  model_outshape = model.output_shape
  img2img = len(model_outshape) > 2
  if len(model_outshape) <= 2:
    nroutputs = model_outshape[-1]
  else:
    model_data_format = get_data_format( model )
    if model_data_format == 'channels_first':
      nroutputs = model_outshape[1]
    else:
      nroutputs = model_outshape[-1]

  res = None
  if withpred:
    if isclassification:
      if not (doprobabilities or withconfidence) and hasattr(model, 'predict_classes'):
        try:
          res = model.predict_classes( samples, batch_size=batch_size )
        except AttributeError:
          pass
    if not isinstance( res, np.ndarray ):
      res = model.predict( samples, batch_size=batch_size )
      res = adaptFromModel(model,res,inp_shape,ret_data_format=data_format)
      ret.update({dgbkeys.preddictstr: res})

  if isclassification and (doprobabilities or withconfidence or withpred):
    if len(ret)<1:
      allprobs = model.predict( samples, batch_size=batch_size )
      allprobs = adaptFromModel(model,allprobs,inp_shape,ret_data_format=data_format)
    else:
      allprobs = ret[dgbkeys.preddictstr]
    indices = None
    if withconfidence or not img2img or (img2img and nroutputs>2):
      N = 2
      if img2img:
        indices = np.argpartition(allprobs,-N,axis=1)[:,-N:]
      else:
        indices = np.argpartition(allprobs,-N,axis=0)[-N:]
    if withpred and isinstance( indices, np.ndarray ):
      if img2img:
        ret.update({dgbkeys.preddictstr: indices[:,-1:]})
      else:
        ret.update({dgbkeys.preddictstr: indices[-1:]})
    if doprobabilities and len(withprobs) > 0:
      res = np.copy(allprobs[withprobs])
      ret.update({dgbkeys.probadictstr: res})
    if withconfidence:
      x = allprobs.shape[-1]
      sortedprobs = allprobs[indices.ravel(),np.tile(np.arange(x),N)].reshape(N,x)
      res = np.diff(sortedprobs,axis=0)
      ret.update({dgbkeys.confdictstr: res})

  return ret


def adaptToModel( model, samples, sample_data_format='channels_first' ):
  nrdims = len( model.input_shape ) - 2
  nrsamples = samples.shape[0]
  samples_nrdims = len(samples.shape)
  model_data_format = get_data_format( model )
  modelcubeszs = getCubeletShape( model )
  if sample_data_format == 'channels_first':
    nrattribs = samples.shape[1]
    cube_shape = samples.shape[2:]
  else:
    nrattribs = samples.shape[-1]
    cube_shape = samples.shape[1:-1]
  shapelims = ()
  idx = 0
  shrinked = False
  for i in cube_shape:
    if i == 1:
      dimsz = 1
    else:
      dimsz = min(i,modelcubeszs[idx])
      if dimsz < i:
        shrinked = True
      idx += 1
    shapelims += (dimsz,)
  cube_shape = np.squeeze( np.empty( shapelims, dtype='uint8' ) ).shape
  datadims = len(cube_shape)
  out_shape = (nrsamples,)
  if model_data_format == 'channels_first':
    out_shape += (nrattribs,)
  for i in cube_shape:
    out_shape += (i,)
  if len(cube_shape) < 1:
    out_shape += (1,)
  if model_data_format == 'channels_last':
    out_shape += (nrattribs,)
  switchedattribs = model_data_format != sample_data_format
  if switchedattribs or nrdims != datadims or shrinked or len(cube_shape) < len(shapelims):
    ret = np.empty( out_shape, dtype=samples.dtype )
    if model_data_format == 'channels_last':
      if nrdims == 3:
        if switchedattribs:
          for iattr in range(nrattribs):
            ret[:,:,:,:,iattr] = np.reshape( samples[:,iattr,:shapelims[0],:shapelims[1],:shapelims[2]], ret[:,:,:,:,iattr].shape )
        else:
          for iattr in range(nrattribs):
            ret[:,:,:,:,iattr] = np.reshape( samples[:,:shapelims[0],:shapelims[1],:shapelims[2],iattr], ret[:,:,:,:,iattr].shape )
      elif nrdims == 2:
        if switchedattribs:
          if samples_nrdims == 5:
            for iattr in range(nrattribs):
              ret[:,:,:,iattr] = np.reshape( np.squeeze( samples[:,iattr,:shapelims[0],:shapelims[1],:shapelims[2]] ), ret[:,:,:,iattr].shape )
          elif samples_nrdims == 4:
            for iattr in range(nrattribs):
              ret[:,:,:,iattr] = np.reshape( np.squeeze( samples[:,iattr,:shapelims[0],:shapelims[1]] ), ret[:,:,:,iattr].shape )
          elif samples_nrdims == 3:
            for iattr in range(nrattribs):
              ret[:,:,:,iattr] = np.reshape( np.squeeze( samples[:,iattr,:shapelims[0]] ), ret[:,:,:,iattr].shape )
        else:
          if samples_nrdims == 5:
            for iattr in range(nrattribs):
              ret[:,:,:,iattr] = np.reshape( np.squeeze( samples[:,:shapelims[0],:shapelims[1],:shapelims[2],iattr] ), ret[:,:,:,iattr].shape )
          elif samples_nrdims == 4:
            for iattr in range(nrattribs):
              ret[:,:,:,iattr] = np.reshape( np.squeeze( samples[:,:shapelims[0],:shapelims[1],iattr] ), ret[:,:,:,iattr].shape )
          elif samples_nrdims == 3:
            for iattr in range(nrattribs):
              ret[:,:,:,iattr] = np.reshape( np.squeeze( samples[:,:shapelims[0],iattr] ), ret[:,:,:,iattr].shape )
      elif nrdims == 1:
        if switchedattribs:
          if samples_nrdims == 5:
            for iattr in range(nrattribs):
              ret[:,:,iattr] = np.reshape( np.squeeze( samples[:,iattr,:shapelims[0],:shapelims[1],:shapelims[2]] ), ret[:,:,iattr].shape )
          elif samples_nrdims == 4:
            for iattr in range(nrattribs):
              ret[:,:,iattr] = np.reshape( np.squeeze( samples[:,iattr,:shapelims[0],:shapelims[1]] ), ret[:,:,iattr].shape )
          elif samples_nrdims == 3:
            for iattr in range(nrattribs):
              ret[:,:,iattr] = np.reshape( samples[:,iattr,:shapelims[0]], ret[:,:,iattr].shape )
        else:
          if samples_nrdims == 5:
            for iattr in range(nrattribs):
              ret[:,:,iattr] = np.reshape( np.squeeze( samples[:,:shapelims[0],:shapelims[1],:shapelims[2],iattr] ), ret[:,:,iattr].shape )
          elif samples_nrdims == 4:
            for iattr in range(nrattribs):
              ret[:,:,iattr] = np.reshape( np.squeeze( samples[:,:shapelims[0],:shapelims[1],iattr] ), ret[:,:,iattr].shape )
          elif samples_nrdims == 3:
            for iattr in range(nrattribs):
              ret[:,:,iattr] = np.reshape( samples[:,:shapelims[0],iattr], ret[:,:,iattr].shape )
      else:
        return samples
      return ret
    else:
      if nrdims == 3:
        if switchedattribs:
          for iattr in range(nrattribs):
            ret[:,iattr] = np.reshape( samples[:,:shapelims[0],:shapelims[1],:shapelims[2],iattr], ret[:,iattr].shape )
        else:
          for iattr in range(nrattribs):
            ret[:,iattr] = np.reshape( samples[:,iattr,:shapelims[0],:shapelims[1],:shapelims[2]], ret[:,iattr].shape )
      elif nrdims == 2:
        if switchedattribs:
          if samples_nrdims == 5:
            for iattr in range(nrattribs):
              ret[:,iattr] = np.reshape( np.squeeze( samples[:,:shapelims[0],:shapelims[1],:shapelims[2],iattr] ), ret[:,iattr].shape )
          elif samples_nrdims == 4:
            for iattr in range(nrattribs):
              ret[:,iattr] = np.reshape( np.squeeze( samples[:,:shapelims[0],:shapelims[1],iattr] ), ret[:,iattr].shape )
          elif samples_nrdims == 3:
            for iattr in range(nrattribs):
              ret[:,iattr] = np.reshape( np.squeeze( samples[:,:shapelims[0],iattr] ), ret[:,iattr].shape )
        else:
          if samples_nrdims == 5:
            for iattr in range(nrattribs):
              ret[:,iattr] = np.reshape( np.squeeze( samples[:,iattr,:shapelims[0],:shapelims[1],:shapelims[2]] ), ret[:,iattr].shape )
          elif samples_nrdims == 4:
            for iattr in range(nrattribs):
              ret[:,iattr] = np.reshape( np.squeeze( samples[:,iattr,:shapelims[0],:shapelims[1]] ), ret[:,iattr].shape )
          elif samples_nrdims == 3:
            for iattr in range(nrattribs):
              ret[:,iattr] = np.reshape( np.squeeze( samples[:,iattr,:shapelims[0]] ), ret[:,iattr].shape )
      elif nrdims == 1:
        if switchedattribs:
          if samples_nrdims == 5:
            for iattr in range(nrattribs):
              ret[:,iattr] = np.reshape( np.squeeze( samples[:,:shapelims[0],:shapelims[1],:shapelims[2],iattr] ), ret[:,iattr].shape )
          elif samples_nrdims == 4:
            for iattr in range(nrattribs):
              ret[:,iattr] = np.reshape( np.squeeze( samples[:,:shapelims[0],:shapelims[1],iattr] ), ret[:,iattr].shape )
          elif samples_nrdims == 3:
            for iattr in range(nrattribs):
              ret[:,iattr] = np.reshape( samples[:,:shapelims[0],iattr], ret[:,iattr].shape )
        else:
          if samples_nrdims == 5:
            for iattr in range(nrattribs):
              ret[:,iattr] = np.reshape( np.squeeze( samples[:,iattr,:shapelims[0],:shapelims[1],:shapelims[2]] ), ret[:,iattr].shape )
          elif samples_nrdims == 4:
            for iattr in range(nrattribs):
              ret[:,iattr] = np.reshape( np.squeeze( samples[:,iattr,:shapelims[0],:shapelims[1]] ), ret[:,iattr].shape )
          elif samples_nrdims == 3:
            for iattr in range(nrattribs):
              ret[:,iattr] = np.reshape( samples[:,iattr,:shapelims[0]], ret[:,iattr].shape )
      else:
        return samples
      return ret
  return samples

def adaptFromModel( model, samples, inp_shape, ret_data_format ):
  nrdims = len( model.output_shape )
  if nrdims == 2:
    return samples.transpose()

  nrpts = inp_shape[0]
  cube_shape = (nrpts,)
  model_data_format = get_data_format( model )
  switchedattribs = model_data_format != ret_data_format
  shapelims = ()
  if model_data_format == 'channels_first':
    nrattribs = model.output_shape[1]
    for i in range(2,nrdims):
      shapelims += (model.output_shape[i],)
  else:
    nrattribs = model.output_shape[-1]
    for i in range(1,nrdims-1):
      shapelims += (model.output_shape[i],)

  data_dims = len(inp_shape)
  if ret_data_format == 'channels_first':
    cube_shape += (nrattribs,)
  if ret_data_format == 'channels_first':
    for i in range(2,data_dims):
      cube_shape += (inp_shape[i],)
  else:
    for i in range(1,data_dims-1):
      cube_shape += (inp_shape[i],)
  if ret_data_format == 'channels_last':
    cube_shape += (nrattribs,)

  res = np.zeros( cube_shape, samples.dtype )
  if model_data_format == 'channels_last':
    if nrdims == 5:
      if switchedattribs:
        for iattr in range(nrattribs):
          res[:,iattr,:shapelims[0],:shapelims[1],:shapelims[2]] = samples[:,:,:,:,iattr]
      else:
        res[:,:shapelims[0],:shapelims[1],:shapelims[2]] = samples
    if nrdims == 4:
      if data_dims == 5:
        if switchedattribs:
          for iattr in range(nrattribs):
            res[:,iattr,:,:shapelims[0],:shapelims[1]] = samples[:,:,:,iattr]
        else:
          res[:,:,:shapelims[0],:shapelims[1]] = samples
      elif data_dims == 4:
        if switchedattribs:
          for iattr in range(nrattribs):
            res[:,iattr,:shapelims[0],:shapelims[1]] = samples[:,:,:,iattr]
        else:
          res[:,:shapelims[0],:shapelims[1]] = samples
    if nrdims == 3:
      if data_dims == 5:
        if switchedattribs:
          for iattr in range(nrattribs):
            res[:,iattr,:,:,:shapelims[0]] = samples[:,:,iattr]
        else:
          res[:,:,:,:shapelims[0]] = samples
      elif data_dims == 4:
        if switchedattribs:
          for iattr in range(nrattribs):
            res[:,iattr,:,:shapelims[0]] = samples[:,:,iattr]
        else:
          res[:,:,:shapelims[0]] = samples
      elif data_dims == 3:
        if switchedattribs:
          for iattr in range(nrattribs):
            res[:,iattr,:shapelims[0]] = samples[:,:,iattr]
        else:
          res[:,:shapelims[0]] = samples
  else:
    if nrdims == 5:
      if switchedattribs:
        for iattr in range(nrattribs):
          res[:,:shapelims[0],:shapelims[1],:shapelims[2],iattr] = samples[:,iattr]
      else:
        res[:,:,:shapelims[0],:shapelims[1],:shapelims[2]] = samples
    if nrdims == 4:
      if data_dims == 5:
        if switchedattribs:
          for iattr in range(nrattribs):
            res[:,:,:shapelims[0],:shapelims[1],iattr] = samples[:,iattr]
        else:
          res[:,:,:,:shapelims[0],:shapelims[1]] = samples
      elif data_dims == 4:
        if switchedattribs:
          for iattr in range(nrattribs):
            res[:,:shapelims[0],:shapelims[1],iattr] = samples[:,iattr]
        else:
          res[:,:,:shapelims[0],:shapelims[1]] = samples
    if nrdims == 3:
      if data_dims == 5:
        if switchedattribs:
          for iattr in range(nrattribs):
            res[:,:,:,:shapelims[0],iattr] = samples[:,iattr]
        else:
          res[:,:,:,:,:shapelims[0]] = samples
      elif data_dims == 4:
        if switchedattribs:
          for iattr in range(nrattribs):
            res[:,:,:shapelims[0],iattr] = samples[:,iattr]
        else:
          res[:,:,:,:shapelims[0]] = samples
      elif data_dims == 3:
        if switchedattribs:
          for iattr in range(nrattribs):
            res[:,:shapelims[0],iattr] = samples[:,iattr]
        else:
          res[:,:,:shapelims[0]] = samples

  return res

def plot( model, outfnm, showshapes=True, withlaynames=False, vertical=True ):
  try:
    import pydot
  except ImportError:
    log_msg( 'Cannot plot the model without pydot module' )
    return
  rankdir = 'TB'
  if not vertical:
    rankdir = 'LR'
  from keras.utils import plot_model
  plot_model( model, to_file=outfnm, show_shapes=showshapes,
              show_layer_names=withlaynames, rankdir=rankdir )

def compute_capability_from_device_desc(device_desc):
  match = re.search(r"compute capability: (\d+)\.(\d+)", \
                      device_desc.physical_device_desc)
  if not match:
    return 0, 0
  return int(match.group(1)), int(match.group(2))

def getDevicesInfo( gpusonly=True ):
  from tensorflow.python.client import device_lib
  local_device_protos = device_lib.list_local_devices()
  if gpusonly:
    ret = [x for x in local_device_protos if x.device_type == 'GPU']
  else:
    ret = [x for x in local_device_protos]
  return ret

def is_gpu_ready():
  import tensorflow as tf
  devs = getDevicesInfo()
  if len(devs) < 1:
    return False
  cc = compute_capability_from_device_desc( devs[0] )
  return tf.test.is_gpu_available(True,cc)

def need_channels_last():
  import tensorflow as tf
  if tf.test.is_built_with_cuda():
    gpudevs = tf.config.get_visible_devices('GPU')
    if len(gpudevs) > 0:
      return False
  return True



#__________________________________________________________________________
#
# (C) dGB Beheer B.V.; (LICENSE) http://opendtect.org/OpendTect_license.txt
# Author:        A. Huck
# Date:          Jan 2019
#
# _________________________________________________________________________
# various tools machine learning using Sci-kit platform
#

import os.path
import sys
import h5py
import json
import joblib
import numpy as np
import pickle
from pathlib import PurePosixPath, PureWindowsPath

import sklearn
from sklearn.preprocessing import StandardScaler
from sklearn import preprocessing
from sklearn.linear_model import LinearRegression, LogisticRegression
from sklearn.ensemble import AdaBoostClassifier, AdaBoostRegressor
from sklearn.ensemble import GradientBoostingClassifier, GradientBoostingRegressor
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor
from sklearn.tree import DecisionTreeRegressor
#from sklearn.multioutput import MultiOutputRegressor
from sklearn.neural_network import MLPClassifier, MLPRegressor
from sklearn.svm import SVC, LinearSVR

from odpy.common import log_msg, redirect_stdout, restore_stdout, isWin
from odpy.oscommand import printProcessTime
import odpy.hdf5 as odhdf5
import dgbpy.keystr as dgbkeys
from dgbpy import hdf5 as dgbhdf5
from multiprocessing import cpu_count

tot_cpu = cpu_count()
n_cpu = tot_cpu

def hasXGBoost():
  try:
    import xgboost
  except ModuleNotFoundError:
    return False
  return True

platform = (dgbkeys.scikitplfnm,'Scikit-learn')
regmltypes = (\
            ('linear','Linear'),\
            ('ensemble','Ensemble'),\
            ('neuralnet','Neural Network'),\
            ('svm','SVM')
          )
classmltypes = (\
            ('logistic','Logistic'),\
            ('ensemble','Ensemble'),\
            ('neuralnet','Neural Network'),\
            ('svm','SVM')
          )
lineartypes = [ ('oslq','Ordinary Least Squares') ]
logistictypes = [ ('log','Logistic Regression Classifier') ]
ensembletypes = []
if hasXGBoost():
  ensembletypes.append( ('xgbdt','XGBoost: (Decision Tree)') )
  ensembletypes.append( ('xgbrf','XGBoost: (Random Forests)') )
ensembletypes.append( ('randfor','Random Forests') )
ensembletypes.append( ('gbc','Gradient Boosting') )
ensembletypes.append( ('ada','Adaboost') )

nntypes = [ ('mlp','Multi-Layer Perceptron') ]
svmtypes = [ ('svm','Support Vector Machine') ]

solvertypes = [\
                ('newton-cg','Newton-CG'),\
                ('lbfgs','Lbfgs'),\
                ('liblinear','Liblinear'),\
                ('sag','Sag'),\
                ('saga','Saga'),\
              ]
kerneltypes = [\
                ('linear','Linear'),\
                ('poly','Polynomial'),\
                ('rbf','Radial Basis Function'),\
                ('sigmoid','Sigmoid'),\
              ]

savetypes = ( 'joblib', 'pickle' )
defsavetype = savetypes[0]

def getMLPlatform():
  return platform[0]

def getUIMLPlatform():
  return platform[1]

def getUiModelTypes(isclassification):
  if isclassification:
    return dgbkeys.getNames( classmltypes )
  return dgbkeys.getNames( regmltypes )

def getUiLinearTypes():
  return dgbkeys.getNames( lineartypes )

def getUiLogTypes():
  return dgbkeys.getNames( logistictypes )

def getUiEnsembleTypes():
  return dgbkeys.getNames( ensembletypes )

def getUiNNTypes():
  return dgbkeys.getNames( nntypes )

def getUiSVMTypes():
  return dgbkeys.getNames( svmtypes)

def getUiSolverTypes():
  return dgbkeys.getNames( solvertypes )

def getUiNNKernelTypes():
  return dgbkeys.getNames( kerneltypes )

def getDefaultSolver( uiname=True ):
  solverstr = LogisticRegression().solver
  return dgbkeys.getNameFromList( solvertypes, solverstr, uiname )

def getDefaultNNKernel( uiname=True ):
  kernelstr = SVC().kernel
  return dgbkeys.getNameFromList( kerneltypes, kernelstr, uiname )


scikit_dict = {
  'ensemblepars': {
    'xgdt': {
      'lr': 1,
      'maxdep': 1,
      'est': 1
      },
    'xgrf': {
      'lr': 1,
      'maxdep': 1,
      'est': 1
      },
    'rf': {
      'maxdep': 50, #default: None, but we prefer less
      'est': RandomForestRegressor().n_estimators
      },
    'gb': {
      'maxdep': GradientBoostingRegressor().max_depth,
      'est': GradientBoostingRegressor().n_estimators,
      'lr': GradientBoostingRegressor().learning_rate
      },
    'ada': {
      'est': AdaBoostRegressor().n_estimators,
      'lr': AdaBoostRegressor().learning_rate
      },
    },
  'nnpars': {
    'maxitr': MLPRegressor().max_iter,
    'lr': MLPRegressor().learning_rate_init,
#    'laysizes': (50,25,5),
    'lay1': 50,
    'lay2': 25,
    'lay3': 5,
    'lay4': 3,
    'lay5': 3,
    'nb': 3
    },
  'svmpars': {
    'kernel': getDefaultNNKernel(False),
    'degree': SVC().degree
    }
}
if hasXGBoost():
  from xgboost import XGBRegressor, XGBRFRegressor
  defdtregressor = XGBRegressor()
  xgdtpars = {
    'lr': scikit_dict['ensemblepars']['xgdt']['lr'],
    'maxdep': scikit_dict['ensemblepars']['xgdt']['maxdep'],
    'est': defdtregressor.n_estimators,
  }
  if defdtregressor.learning_rate != None:
    xgdtpars.update({'lr': defdtregressor.learning_rate})
  if defdtregressor.max_depth != None:
    xgdtpars.update({'maxdep': defdtregressor.max_depth})
  scikit_dict['ensemblepars'].update({'xgdt': xgdtpars})

  defrfregressor = XGBRFRegressor()
  xgrfpars = {
      'lr': defrfregressor.learning_rate,
      'maxdep': scikit_dict['ensemblepars']['xgrf']['maxdep'],
      'est': defrfregressor.n_estimators,
  }
  if defrfregressor.max_depth != None:
    xgrfpars.update({'maxdep': defrfregressor.max_depth})
  scikit_dict['ensemblepars'].update({'xgrf': xgrfpars})

def getLinearPars( modelname='Ordinary Least Squares'):
  return {
    'decimation': False,
    'modelname': modelname
    }

def getLogPars( modelname='Logistic Regression Classifier',solver=None):
  if solver == None:
    solver = getDefaultSolver()
  return {
    'decimation': False,
    'modelname': modelname,
    'solver': solver
    }

def getEnsembleParsXGDT( modelname='XGBoost: (Decision Tree)',
                       maxdep=scikit_dict['ensemblepars']['xgdt']['maxdep'],
                       est=scikit_dict['ensemblepars']['xgdt']['est'],
                       lr=scikit_dict['ensemblepars']['xgdt']['lr'] ):
  return {
    'decimation': False,
    'maxdep' : maxdep,
    'est': est,
    'lr': lr,
    'modelname': modelname
    }

def getEnsembleParsXGRF( modelname='XGBoost: (Random Forests)',
                       maxdep=scikit_dict['ensemblepars']['xgrf']['maxdep'],
                       est=scikit_dict['ensemblepars']['xgrf']['est'],
                       lr=scikit_dict['ensemblepars']['xgrf']['lr'] ):
  return {
    'decimation': False,
    'maxdep' : maxdep,
    'est': est,
    'lr': lr,
    'modelname': modelname
    }

def getEnsembleParsRF( modelname='Random Forests',
                       maxdep=scikit_dict['ensemblepars']['rf']['maxdep'],
                       est=scikit_dict['ensemblepars']['rf']['est'] ):
  return {
    'decimation': False,
    'maxdep' : maxdep,
    'est': est,
    'modelname': modelname
    }

def getEnsembleParsGB( modelname='Gradient Boosting',
                       maxdep=scikit_dict['ensemblepars']['gb']['maxdep'],
                       est=scikit_dict['ensemblepars']['gb']['est'],
                       lr=scikit_dict['ensemblepars']['gb']['lr'] ):
  return {
    'decimation': False,
    'maxdep' : maxdep,
    'est': est,
    'lr': lr,
    'modelname': modelname
    }

def getEnsembleParsAda( modelname='Adaboost',
                        est=scikit_dict['ensemblepars']['ada']['est'],
                        lr=scikit_dict['ensemblepars']['ada']['lr'] ):
  return {
    'decimation': False,
    'est': est,
    'lr': lr,
    'modelname': modelname
    }

def getNNPars( modelname='Multi-Layer Perceptron',
               maxitr = scikit_dict['nnpars']['maxitr'],
               lr = scikit_dict['nnpars']['lr'],
               lay1=scikit_dict['nnpars']['lay1'],
               lay2=scikit_dict['nnpars']['lay2'],
               lay3=scikit_dict['nnpars']['lay3'],
               lay4=scikit_dict['nnpars']['lay4'],
               lay5=scikit_dict['nnpars']['lay5'],
               nb = scikit_dict['nnpars']['nb']):
  return {
    'decimation': False,
    'maxitr': maxitr,
    'lr': lr,
    'lay1': lay1,
    'lay2': lay2,
    'lay3': lay3,
    'lay4': lay4,
    'lay5': lay5,
    'nb': nb,
    'modelname': modelname
    }

def getSVMPars( modelname='Support Vector Machine',
                kernel = scikit_dict['svmpars']['kernel'],
                degree = scikit_dict['svmpars']['degree'] ):
  return {
    'decimation': False,
    'kernel': kernel,
    'degree': degree,
    'modelname': modelname
    }

def getNewScaler( mean, scale ):
  scaler = StandardScaler()
  scaler.mean_ = np.array( mean )
  scaler.scale_ = np.array( scale )
  scaler.var_ = np.square( scaler.scale_ )
  scaler.n_samples_seen_ = len(mean)
  return scaler

def getScaler( x_train, byattrib ):
  nrattribs = x_train.shape[1]
  mean = list()
  var = list()
  if byattrib:
    for a in range(nrattribs):
      inp = x_train[:,a,:,:,:]
      mean.append( np.mean(inp,dtype=np.float64) )
      var.append( np.var(inp,dtype=np.float64) )
  else:
    mean.append( np.mean(x_train,dtype=np.float64) )
    var.append( np.var(x_train,dtype=np.float64) )
  scaler = StandardScaler()
  scaler.mean_ = np.array( mean )
  scaler.var_ = np.array( var )
  scaler.scale_ = np.sqrt( scaler.var_ )
  scaler.n_samples_seen_ = len(mean)
  return scaler

def transform( samples, mean, stddev ):
  samples -= mean
  samples /= stddev
  return samples

def transformBack( samples, mean, stddev ):
  samples *= stddev
  samples += mean
  return samples

def scale( samples, scaler ):
  if scaler == None:
    return samples
  if scaler.n_samples_seen_ == 1:
    samples = transform( samples, scaler.mean_[0], scaler.scale_[0] )
  else:
    for i in range(scaler.n_samples_seen_):
      samples[:,i] = transform( samples[:,i], scaler.mean_[i], scaler.scale_[i] )
  
  return samples

def unscale( samples, scaler ):
  if scaler == None:
    return samples
  if scaler.n_samples_seen_ == 1:
    samples = transformBack( samples, scaler.mean_[0], scaler.scale_[0] )
  else:
    for i in range(scaler.n_samples_seen_):
      samples[:,i] = transformBack( samples[:,i], scaler.mean_[i], scaler.scale_[i] )
  
  return samples

def getDefaultModel( setup, params=scikit_dict ):
  modelname = params['modelname']
  isclassification = setup[dgbhdf5.classdictstr]
  try:
    if modelname == 'Ordinary Least Squares':
      model = LinearRegression(n_jobs=-1)
    elif modelname == 'Logistic Regression Classifier':
      solvernm = dgbkeys.getNameFromUiName( solvertypes, params['solver'] )
      model = LogisticRegression(solver=solvernm,n_jobs=-1)
    elif modelname == 'XGBoost: (Decision Tree)':
      from xgboost import XGBClassifier, XGBRegressor
      learning_rate = params['lr']
      max_depth = params['maxdep']
      n_estimators = params['est']
      if isclassification:
        model = XGBClassifier(n_estimators=n_estimators,max_depth=max_depth,learning_rate=learning_rate,n_jobs=n_cpu)
      else:
        model = XGBRegressor(objective='reg:squarederror',n_estimators=n_estimators,max_depth=max_depth,learning_rate=learning_rate,n_jobs=n_cpu)
    elif modelname == 'XGBoost: (Random Forests)':
      from xgboost import XGBRFClassifier, XGBRFRegressor
      learning_rate = params['lr']
      max_depth = params['maxdep']
      n_estimators = params['est']
      if isclassification:
        model = XGBRFClassifier(n_estimators=n_estimators,max_depth=max_depth,learning_rate=learning_rate,n_jobs=n_cpu)
      else:
        model = XGBRFRegressor(objective='reg:squarederror',n_estimators=n_estimators,max_depth=max_depth,learning_rate=learning_rate,n_jobs=n_cpu)
    elif modelname == 'Random Forests':
      n_estimators = params['est']
      max_depth = params['maxdep']
      if isclassification:
        model = RandomForestClassifier(n_estimators=n_estimators,criterion='gini',max_depth=max_depth,n_jobs=-1)
      else:
        model = RandomForestRegressor(n_estimators=n_estimators,criterion='mse',max_depth=max_depth,n_jobs=-1)
    elif modelname == 'Gradient Boosting':
      n_estimators = params['est']
      learning_rate = params['lr']
      max_depth = params['maxdep']
      if isclassification:
        model = GradientBoostingClassifier(learning_rate=learning_rate,n_estimators=n_estimators,max_depth=max_depth)
      else:
        model = GradientBoostingRegressor(learning_rate=learning_rate,n_estimators=n_estimators,max_depth=max_depth)
    elif modelname == 'Adaboost':
      n_estimators = params['est']
      learning_rate = params['lr']
      if isclassification:
        model = AdaBoostClassifier(n_estimators=n_estimators,learning_rate=learning_rate)
      else:
        model = AdaBoostRegressor(n_estimators=n_estimators,learning_rate=learning_rate)
    elif modelname == 'Multi-Layer Perceptron':
      lay1 = params['lay1']
      lay2 = params['lay2']
      lay3 = params['lay3']
      lay4 = params['lay4']
      lay5 = params['lay5']
      nb = params['nb']
      hidden_layer = (lay1,lay2,lay3,lay4,lay5)
      hidden_layer = hidden_layer[0:nb]
      learning_rate = params['lr']
      max_iter = params['maxitr']
      if isclassification:
        model = MLPClassifier(hidden_layer_sizes=hidden_layer,learning_rate_init=learning_rate,max_iter=max_iter)
      else:
        model = MLPRegressor(hidden_layer_sizes=hidden_layer,learning_rate_init=learning_rate,max_iter=max_iter)
    elif modelname == 'Support Vector Machine':
      kernel = dgbkeys.getNameFromUiName( kerneltypes, params['kernel'] )
      degree = params['degree']
      if isclassification:
        model = SVC(kernel=kernel,degree=degree)
      else:
        model = LinearSVR(kernel=kernel,degree=degree)
  except Exception as e:
    log_msg( 'Exception:', e )
    raise e
  return model

def train(model, trainingdp):
  x_train = trainingdp[dgbkeys.xtraindictstr]
  y_train = trainingdp[dgbkeys.ytraindictstr].ravel()
  printProcessTime( 'Training with scikit-learn', True, print_fn=log_msg )
  log_msg( '\nTraining on', len(y_train), 'samples' )
  log_msg( 'Validate on', len(trainingdp[dgbkeys.yvaliddictstr]), 'samples\n' )
  redirect_stdout()
  model.verbose = 51
  ret = model.fit(x_train,y_train)
  restore_stdout()
  printProcessTime( 'Training with scikit-learn', False, print_fn=log_msg, withprocline=False )
  assessQuality( model, trainingdp )
  return ret

def assessQuality( model, trainingdp ):
  if not dgbkeys.yvaliddictstr in trainingdp:
    return
  try:
    x_validate = trainingdp[dgbkeys.xvaliddictstr]
    y_validate = trainingdp[dgbkeys.yvaliddictstr].ravel()
    y_predicted = model.predict(x_validate)
    if trainingdp[dgbkeys.infodictstr][dgbkeys.classdictstr]:
      cc = np.sum( y_predicted==y_validate) / len(y_predicted)
    else:
      cc = np.corrcoef( y_predicted, y_validate )[0,1]
    log_msg( '\nCorrelation coefficient with validation data: ', "%.4f" % cc, '\n' )
  except Exception as e:
    log_msg( '\nCannot compute model quality:' )
    log_msg( repr(e) )

def save( model, outfnm, save_type=defsavetype ):
  h5file = odhdf5.openFile( outfnm, 'w' )
  odhdf5.setAttr( h5file, 'backend', 'scikit-learn' )
  odhdf5.setAttr( h5file, 'sklearn_version', sklearn.__version__ )
  odhdf5.setAttr( h5file, 'type', 'RandomForestRegressor' )
  odhdf5.setAttr( h5file, 'model_config', json.dumps(model.get_params()) )
  modelgrp = h5file.create_group( 'model' )
  odhdf5.setAttr( modelgrp, 'type', save_type )
  if save_type == savetypes[0]:
    joutfnm = os.path.splitext( outfnm )[0] + '.joblib'
    joblib.dump( model, joutfnm )
    odhdf5.setAttr( modelgrp, 'path', joutfnm )
  elif save_type == savetypes[1]:
    exported_modelstr = pickle.dumps(model)
    exported_model = np.frombuffer( exported_modelstr, dtype='S1', count=len(exported_modelstr) )
    modelgrp.create_dataset('object',data=exported_model)
  h5file.close()

def translateFnm( modfnm, modelfnm ):
  posidxh5fp = PurePosixPath( modelfnm )
  winh5fp = PureWindowsPath( modelfnm )
  posixmodfp = PurePosixPath( modfnm )
  winmodfp = PureWindowsPath( modfnm )
  if isWin():
    moddir = winh5fp.parent
    modbasefnm = winmodfp.name
    modlocfnm = PureWindowsPath( moddir ).joinpath( PureWindowsPath(modbasefnm))
    if os.path.exists(modlocfnm):
      modfnm = modlocfnm
    else:
      moddir = posidxh5fp.parent
      modbasefnm = posixmodfp.name
      modlocfnm = PurePosixPath( moddir ).joinpath( PurePosixPath(modbasefnm) )
      if os.path.exists(modlocfnm):
        modfnm = modlocfnm
  else:
    moddir = posidxh5fp.parent
    modbasefnm = posixmodfp.name
    modlocfnm = PurePosixPath( moddir ).joinpath( PurePosixPath(modbasefnm) )
    if os.path.exists(modlocfnm):
      modfnm = modlocfnm
    else:
      moddir = winh5fp.parent
      modbasefnm = winmodfp.name
      modlocfnm = PureWindowsPath( moddir).joinpath(PureWindowsPath(modbasefnm))
      modlocfnm = modlocfnm.as_posix()
      if os.path.exists(modlocfnm):
        modfnm = modlocfnm
  return modfnm

def load( modelfnm ):
  model = None
  h5file = odhdf5.openFile( modelfnm, 'r' )

  modelpars = json.loads( odhdf5.getAttr(h5file,'model_config') )
  modeltype = odhdf5.getText( h5file, 'type' )
  info = odhdf5.getInfoDataSet( h5file )

  modelgrp = h5file['model']
  savetype = odhdf5.getText( modelgrp, 'type' )
  if savetype == savetypes[0]:
    modfnm = odhdf5.getText( modelgrp, 'path' )
    modfnm = translateFnm( modfnm, modelfnm )
    model = joblib.load( modfnm )
  elif savetype == savetypes[1]:
    modeldata = modelgrp['object']
    model = pickle.loads( modeldata[:].tostring() )

  h5file.close()
  return model
 
def apply( model, samples, scaler, isclassification, withpred, withprobs, withconfidence, doprobabilities ):
  model.verbose = 0
  samples = np.reshape( samples, (len(samples),-1) )
  if scaler != None:
    samples = scaler.transform( samples )

  ret = {}
  res = None
  if withpred:
    res = np.transpose( model.predict( samples ) )
    ret.update({dgbkeys.preddictstr: res})

  if isclassification and (doprobabilities or withconfidence):
    res = np.transpose( model.predict_proba( samples ) )
    ret.update({dgbkeys.probadictstr: res})

  return ret



import sys
import os
import collections
import numpy as np
import h5py
import odpy.hdf5 as odhdf5
import odpy.ranges as ranges
import odpy.volpreload as volpreload
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1.inset_locator import inset_axes

if len(sys.argv) < 2:
  print( "No file name specified" )

filenm = sys.argv[1]
filebasenm = os.path.basename( filenm )
print( "\nBrowsing '" + filebasenm + "'\n\n" )

si = odhdf5.getSurveyInfo( filenm )
attribinfo = odhdf5.getAttribInfo( filenm )
samplinginfo = attribinfo['range']
datatkzs = ranges.getAxesAsRanges( samplinginfo )
axes = ranges.getAxesAsArrays( samplinginfo )
blocksinfo = attribinfo['block']
blocksrg = blocksinfo['range']

print( "Data is stored as " + attribinfo['storage'] + "\n" )
# TODO: add gap information: The cube has varying cross-line ranges. / The cube is 100% rectangular.

usrinp = "1"
compnms = attribinfo['attributes']
attribnm = None
if isinstance( compnms, list ):
  nrcomps = len( compnms )
  if nrcomps > 1:
    for idx, compnm in enumerate(compnms):
      print( "Component " + str(idx+1) + ": '" + compnm + "'" )
      idx = idx+1
    usrinp = input( "Enter component index: " )
  cidx = int( usrinp ) - 1
  if cidx < 0:
    cidx = 0
  if cidx >= nrcomps:
    cidx = nrcomps - 1
  attribnm = compnms[cidx]

ranges.printSampling( samplinginfo )
print( "Number of Z samples: " + repr(len(axes['zsamp'])) )
print( "" )

cont = "Y"
while (cont == "Y") or (cont == "y" ) or (cont == "YES") or (cont == "Yes"):
  slicesel = -1
  while (slicesel != 1) and (slicesel != 2) and (slicesel != 3):
    slicesel = input("Slice type: [1-In-line] [2-Cross-line] [3-Z slice]  ")
    try:
      slicesel = int(slicesel)
    except ValueError:
      cont="No"
      break
    inlinedir = slicesel == 1
    crosslinedir = slicesel == 2
    zdir = slicesel == 3
    if inlinedir:
      slicetxt = "Inline"
      slicerg = ranges.getLineObj( datatkzs )
      tkzs = collections.OrderedDict({
        'Crossline': ranges.getTraceObj( datatkzs ),
        'Z': ranges.getZObj( datatkzs )
      })
    elif crosslinedir:
      slicetxt = "Crossline"
      slicerg = ranges.getTraceObj( datatkzs )
      tkzs = collections.OrderedDict({
        'Inline': ranges.getLineObj( datatkzs ),
        'Z': ranges.getZObj( datatkzs )
      })
    elif zdir:
      slicetxt = "Z slice"
      slicerg = ranges.getZObj( datatkzs )
      tkzs = collections.OrderedDict({
        'Inline': ranges.getLineObj( datatkzs ),
        'Crossline': ranges.getTraceObj( datatkzs )
      })
    elif slicesel == 0:
      cont="No"
      break
    else:
      continue
  if cont == "No":
    continue

  slicestep = slicerg.step
  slicenb = input( slicetxt + " number: " )
  sliceidx = 0
  try:
    slicenb = int(slicenb)
  except ValueError:
    print( "Please enter an integer\n" )
    continue
  if slicenb % slicestep != 0:
    print( slicetxt + " " + slicenb + " is not a multiple of the step: " + repr(slicerg.start) + "-" + repr(slicerg.stop) + " (step " + repr(slicestep) + ")" + "\n")
    continue
  try:
    sliceidx = slicerg.index( slicenb )
  except ValueError:
    print( slicetxt + " " + slicenb + "is not in range: " + repr(slicerg.start) + "-" + repr(slicerg.stop) + " (step" + repr(slicestep) + ")" + "\n")
    continue

  if slicesel == 1:
    step = ranges.getLineObj( datatkzs ).step
    tkzs.update({'Inline': range(slicenb,slicenb+step,step)})
    tkzs.move_to_end('Inline',last=False)
  elif slicesel == 2:
    step = ranges.getTraceObj( datatkzs ).step
    tkzs.update({'Crossline': range(slicenb,slicenb+step,step)})
    tkzs.move_to_end('Z',last=True)
  elif slicesel == 3:
    step = ranges.getZObj( datatkzs ).step
    tkzs.update({'Z': range(slicenb,slicenb+step,step)})

  preloaded = volpreload.preLoad( filenm, attribute=attribnm, sampling=tkzs )
  arr = preloaded['data']
  if len(arr) == 0:
    exit(1)

  fig,ax = plt.subplots(1,1,sharex="col",sharey="row",facecolor="white")
  fig.subplots_adjust(top=0.95,bottom=0.05,left=0.05,right=0.93,hspace=0.02,wspace=0.15)
  plt.get_current_fig_manager().window.setGeometry(1920,0,1920,960)
  #TODO: Assign amplitude range from data
  axes = preloaded['axes']
  axrg = axes['range']
  im = ax.imshow(np.transpose(arr),cmap="seismic_r",aspect="auto",interpolation="quadric",vmin=-8000,vmax=8000,extent=[axrg[0][0],axrg[0][-1],axrg[1][-1],axrg[1][0]])
  title = "HDF5 Slice of "+filebasenm
  if attribnm != None:
    title = title + ' [' + attribnm + ']'
  labels = axes['labels']
  ax.set_title(title)
  ax.set_xlabel(labels[0])
  ax.set_ylabel(labels[1])
  axins = inset_axes(ax,width="1%",height="100%",loc=3,bbox_to_anchor=(1.02,0.,1,1),bbox_transform=ax.transAxes,borderpad=0)
  cbar = ax.figure.colorbar(im,cax=axins)
  print( "Close the plot to continue" )
  plt.show()

  cont = input( "Continue [Y/N] (Yes)? " )
  if not cont:
    cont = "Yes"
  print( "" )

h5file = odhdf5.openFile( filenm, 'r' )
#cubeposidx = list( h5file.keys() ).index("Seismic Cube Positions")
#cubepos = list(h5file.values())[cubeposidx]
cubepos = h5file["Seismic Cube Positions"]
h5file.close()

exit( 0 )



#
# (C) dGB Beheer B.V.; (LICENSE) http://opendtect.org/OpendTect_license.txt
# AUTHOR   : A. Huck
# DATE     : June 2019
#
# Deep learning apply server
#
#

import io
import json
import numpy as np
import os
import psutil
import selectors
import struct
import sys
import traceback as tb

from odpy.common import *
import dgbpy.keystr as dgbkeys
from dgbpy import hdf5 as dgbhdf5
from dgbpy import mlio as dgbmlio
from dgbpy import mlapply as dgbmlapply
from dgbpy import dgbscikit, dgbkeras

class ExitCommand(Exception):
    pass

class ModelApplier:
    def __init__(self, modelfnm,isfake=False):
        self.pars_ = None
        self.fakeapply_ = isfake
        self.scaler_ = None
        self.extscaler_ = None
        self.info_ = self._get_info(modelfnm)
        self.model_ = None
        self.applyinfo_ = None
        self.batchsize_ = None
        self.debugstr = ''

    def _get_info(self,modelfnm):
        info = dgbmlio.getInfo( modelfnm, quick=True )
        if self.fakeapply_:
            info[dgbkeys.plfdictstr] = dgbkeys.numpyvalstr
        return info
    
    def setOutputs(self, outputs):
        if self.fakeapply_:
            self.applyinfo_ = dgbmlio.getApplyInfo( self.info_ )
        else:
            self.applyinfo_ = dgbmlio.getApplyInfo( self.info_, outputs )
        (self.scaler_,self.extscaler_) = self.getScaler( outputs )
        if dgbkeras.prefercpustr in outputs:
            dgbkeras.set_compute_device( outputs[dgbkeras.prefercpustr] )
        if dgbkeras.defbatchstr in outputs:
            self.batchsize_ = outputs[dgbkeras.defbatchstr]
        if self.fakeapply_:
            return None
        modelfnm = self.info_[dgbkeys.filedictstr]
        (self.model_,self.info_) = dgbmlio.getModel( modelfnm, fortrain=False )

    def _usePar(self, pars):
        self.pars_ = pars

    def hasModel(self):
        return self.model_ != None

    def getDefaultScaler(self):
        means = list()
        stddevs = list()
        for i in range(dgbhdf5.getNrAttribs(self.info_)):
            stddevs.append( 50 )
            means.append( 128 )
        self.extscaler_ = dgbscikit.getNewScaler( means, stddevs )
        return self.extscaler_

    def getScaler( self, outputs ):
        if not 'scales' in outputs:
            return (self.scaler_,self.extscaler_)

        scales = outputs['scales']
        means = list()
        stddevs = list()
        scaleratios = list()
        for scl in scales:
            means.append( scl['avg'] )
            stddevs.append( scl['stdev'] )
            scaleratios.append( scl['scaleratio'] )

        if len(means) > 0:
            self.scaler_ = dgbscikit.getNewScaler( means, stddevs )
        inputs = self.info_[dgbkeys.inputdictstr]
        if dgbhdf5.isLogInput( self.info_ ):
            inputs = self.info_[dgbkeys.inputdictstr]
            firstinpnm = next(iter(inputs))
            if firstinpnm in inputs:
                inp = inputs[firstinpnm]
                if dgbkeys.scaledictstr in inp:
                    self.scaler_ = inp[dgbkeys.scaledictstr]
        elif outputs[dgbkeys.surveydictstr] in inputs:
            survdirnm = outputs[dgbkeys.surveydictstr]
            inp = inputs[survdirnm]
            if dgbkeys.scaledictstr in inp:
                inpscale = inp[dgbkeys.scaledictstr]
                if dgbkeys.collectdictstr in inp:
                  attribs = inp[dgbkeys.collectdictstr]
                  for scl in scales:
                      applykey = scl[dgbkeys.dbkeydictstr]
                      applynm = scl[dgbkeys.namedictstr]
                      iattr = 0
                      for attribnm in attribs:
                          attrib = attribs[attribnm]
                          if attrib[dgbkeys.dbkeydictstr] == applykey or \
                             attribnm == applynm:
                             idx = attrib[dgbkeys.iddictstr]
                             self.scaler_.scale_[idx] = inpscale.scale_[idx]
                             self.scaler_.mean_[idx] = inpscale.mean_[idx]
                             break
                          else:
                            self.scaler_.scale_[iattr] *= scaleratios[iattr]
                          iattr += 1
                else:
                  for i in range(len(inpscale.scale_)):
                    means.append( inpscale.mean_[i] )
                    stddevs.append( inpscale.scale_[i] )
                  if len(means) > 0:
                    self.scaler_ = dgbscikit.getNewScaler( means, stddevs )
        elif dgbkeys.mlsoftkey in inputs:
            inp = inputs[dgbkeys.mlsoftkey]
            means = list()
            stddevs = list()
            if dgbkeys.scaledictstr in inp:
                inpscale = inp[dgbkeys.scaledictstr]
                for (scale,mean) in zip(inpscale.scale_,inpscale.mean_):
                    stddevs.append( scale )
                    means.append( mean )
                self.extscaler_ = dgbscikit.getNewScaler( means, stddevs )
            else:
                self.extscaler_ = self.getDefaultScaler()

        return (self.scaler_,self.extscaler_)

    def doWork(self,inp):
        nrattribs = inp.shape[0]
        inpshape = self.info_[dgbkeys.inpshapedictstr]
        nrzin = inp.shape[-1]
        vertical =  isinstance(inpshape,int)
        is2d = False
        if vertical:
            chunksz = 1
            nrzoutsamps = nrzin-inpshape+1
        else:
            is2d = len(inp.shape) == 3
            if is2d:
                chunksz = inp.shape[1] - inpshape[1] + 1
            else:
                chunksz = inp.shape[2] - inpshape[1] + 1
            nrzoutsamps = nrzin - inpshape[2] +1
        nroutsamps = nrzoutsamps * chunksz
        samples_shape = dgbhdf5.get_np_shape( inpshape, nrattribs=nrattribs,
                                              nrpts=nrzoutsamps )
        nrtrcs = samples_shape[-2]
        nrz = samples_shape[-1]
        allsamples = list()
        for i in range(chunksz):
          if nrz == 1:
            inp = np.transpose( inp )
            if chunksz < 2:
              allsamples.append( np.resize( np.array(inp), samples_shape ) )
            else:
              allsamples.append( np.resize( np.array(inp), samples_shape ) ) #review
          else:
            loc_samples = np.empty( samples_shape, dtype=inp.dtype )
            if vertical:
              for zidz in range(nrzoutsamps):
                loc_samples[zidz,:,0,0,:] = inp[:,zidz:zidz+nrz]
            else:
              if is2d:
                for zidz in range(nrzoutsamps):
                  loc_samples[zidz] = inp[:,i:i+nrtrcs+1,zidz:zidz+nrz]                 
              else:
                for zidz in range(nrzoutsamps):
                  loc_samples[zidz] = inp[:,:,i:i+nrtrcs+1,zidz:zidz+nrz]
            allsamples.append( loc_samples )
        samples = np.concatenate( allsamples )
#        self.debugstr = self.debug_msg( samples[0,0,0,0,:1].squeeze() )
        samples = dgbscikit.scale( samples, self.scaler_ )
#        self.debugstr = self.debug_msg( samples[0,0,0,0,:1].squeeze() )
        samples = dgbscikit.unscale( samples, self.extscaler_ )
#        self.debugstr = self.debug_msg( samples[0,0,0,0,:1].squeeze() )
#        min = np.min( samples ) 
#        samples = samples-min
#        max = np.max( samples )
#        samples = samples/max
#        samples = samples*255 
        ret = dgbmlapply.doApply( self.model_, self.info_, samples, \
                                  scaler=None, applyinfo=self.applyinfo_, \
                                  batchsize=self.batchsize_ )
        res = list()
        outkeys = list()
        outkeys.append( dgbkeys.preddictstr )
        outkeys.append( dgbkeys.probadictstr )
        outkeys.append( dgbkeys.confdictstr )
        for outkey in outkeys:
          if outkey in ret:
            if chunksz > 1:
              nrattrret = ret[outkey].shape[-1]
              ret[outkey] = np.resize( ret[outkey], (nrzoutsamps,chunksz,nrattrret))
            res.append( ret[outkey] )
        return res

    def debug_msg(self,a,b=None,c=None,d=None,e=None,f=None,g=None,h=None):
        ret = str(a)
        if b != None:
          ret += ' '+str(b)
        if c != None:
          ret += ' '+str(c)
        if d != None:
          ret += ' '+str(d)
        if e != None:
          ret += ' '+str(e)
        if f != None:
          ret += ' '+str(f)
        if g != None:
          ret += ' '+str(g)
        if h != None:
          ret += ' '+str(h)
        if len(self.debugstr) > 0:
          self.debugstr += '\n'
        self.debugstr += ret
        return self.debugstr


class Message:
    def __init__(self, selector, sock, addr, applier):
        self.selector = selector
        self.sock = sock
        self.addr = addr
        self._recv_buffer = b""
        self._send_buffer = b""
        self._payload_len = None
        self._reqid = None
        self._subid = None
        self._jsonheader_len = None
        self.jsonheader = None
        self.request = None
        self.response_created = False
        self.applier = applier
        self.lastmessage = False

    def _set_selector_events_mask(self, mode):
        """Set selector to listen for events: mode is 'r', 'w', or 'rw'."""
        if mode == "r":
            events = selectors.EVENT_READ
        elif mode == "w":
            events = selectors.EVENT_WRITE
        elif mode == "rw":
            events = selectors.EVENT_READ | selectors.EVENT_WRITE
        else:
            raise ValueError(f"Invalid events mask mode {repr(mode)}.")
        self.selector.modify(self.sock, events, data=self)

    def _read(self):
        try:
            # Should be ready to read
            data = self.sock.recv(16777216)
        except BlockingIOError:
            # Resource temporarily unavailable (errno EWOULDBLOCK)
            pass
        else:
            if data:
                self._recv_buffer += data
            else:
                raise RuntimeError("Peer closed.")

    def _write(self):
        if self._send_buffer:
            try:
                # Should be ready to write
                sent = self.sock.send(self._send_buffer)
            except BlockingIOError:
                # Resource temporarily unavailable (errno EWOULDBLOCK)
                pass
            else:
                self._send_buffer = self._send_buffer[sent:]
                # Close when the buffer is drained. The response has been sent.
                if sent and not self._send_buffer:
                    self.close()

    def _json_encode(self, obj, encoding):
        json_hdr = json.dumps(obj, ensure_ascii=False).encode(encoding)
        return struct.pack('=i',len(json_hdr)) + json_hdr

    def _json_decode(self, json_bytes, encoding):
        hdrlen = 4
        json_hdr = struct.unpack('=i',json_bytes[:hdrlen])[0]
        tiow = io.TextIOWrapper(
            io.BytesIO(json_bytes[hdrlen:hdrlen+json_hdr]), encoding=encoding, newline=""
        )
        obj = json.load(tiow)
        tiow.close()
        return (json_hdr,obj,json_bytes[hdrlen+json_hdr:])

    def _array_decode(self, arrptr, shapes, dtypes):
        offset = 0
        arrs = list()
        for shape,dtype in zip(shapes,dtypes):
          nrsamples = np.prod(shape,dtype=np.int64)
          arr = np.frombuffer(arrptr,dtype=dtype,count=nrsamples,offset=offset)
          arr = arr.reshape( shape )
          offset += arr.nbytes
          arrs.append( arr )
        return {
          'action': 'apply',
          'data': arrs
        }

    def _create_message(
        self, *, content_bytes, content_type, content_encoding, arrsize
    ):
        jsonheader = {
            "byteorder": sys.byteorder,
            "content-type": content_type,
            "content-encoding": content_encoding,
            "content-length": len(content_bytes),
        }
        if arrsize != None:
          jsonheader.update({ 'array-shape': arrsize })
        (self,jsonheader) = self._add_debug_str( jsonheader )
        jsonheader_bytes = self._json_encode(jsonheader, 'utf-8')
        payload = jsonheader_bytes + content_bytes
        od_hdr =   struct.pack('=i',len(payload)) \
                 + struct.pack('=i',self._reqid) \
                 + struct.pack('=h',self._subid)
        return od_hdr + payload

    def _create_response_json_content(self):
        action = self.request.get('action')
        content = { 'result': None }
        if action == 'status':
            content['result'] = 'Server online'
            content['pid'] = psutil.Process().pid
        elif action == 'kill':
            content['result'] = 'Kill request received'
            self.lastmessage = True
        elif action == 'outputs':
            try:
              self.applier.setOutputs( self.request.get('value') )
              content['result'] = 'Output names received'
            except Exception as e:
              content = {"result": f'start error exception: {repr(e)}.'}
        else:
            content['result'] = f'Error: invalid action "{action}".'
        content_encoding = 'utf-8'
        response = {
            'content_bytes': self._json_encode(content, content_encoding),
            'content_type': 'text/json',
            'content_encoding': content_encoding,
            'arrsize': None,
        }
        return response

    def _create_response_array_content(self):
        action = self.request.get('action')
        try:
            res = list()
            if action == 'apply':
                for arr in self.request.get('data'):
                    res = self.applier.doWork(arr)
            else:
                content = {"result": f'Error: invalid action "{action}".'}
        except Exception as e:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
            stackstr = ''.join(tb.extract_tb(exc_tb,limit=10).format())
            content = {'result': f'Apply error exception:\n{repr(e)+" on line "+str(exc_tb.tb_lineno)+" of script "+fname}\n{stackstr}\n\n{self.applier.debugstr}'}
            self.applier.debugstr = ''
            content_encoding = 'utf-8'
            response = {
                'content_bytes': self._json_encode(content, content_encoding),
                'content_type': 'text/json',
                'content_encoding': content_encoding,
                'arrsize': None,
            }
            return (self,response)

        ret = bytes()
        dtypes = list()
        shapes = list()
        for arr in res:
          ret += arr.tobytes()
          shapes.append( arr.shape )
          dtypes.append( arr.dtype.name )
        response = {
          'content_bytes': ret,
          'content_type': "binary/array",
          'content_encoding': dtypes,
          'arrsize': shapes,
        }
        return (self,response)

    def _create_response_binary_content(self):
        response = {
            "content_bytes": b"First 10 bytes of request: "
            + self.request[:10],
            "content_type": "binary/custom-server-binary-type",
            "content_encoding": "binary",
            'arrsize': None,
        }
        return response

    def _add_debug_str( self, response ):
        if self.applier == None:
            return (self,response)
        debugstr = self.applier.debugstr
        if len(debugstr) > 0:
            response.update( {'debug-message': debugstr} )
            self.applier.debugstr = ''
        return (self,response)

    def process_events(self, mask):
        if mask & selectors.EVENT_READ:
            self.read()
        if mask & selectors.EVENT_WRITE:
            self.write()

    def read(self):
        self._read()

        if self._payload_len is None:
            self.process_odheader()

        if self._payload_len is not None:
            if self.jsonheader is None:
                self.process_jsonheader()

        if self.jsonheader:
            if self.request is None:
                self.process_request()

    def write(self):
        if self.request:
            if not self.response_created:
                self.create_response()

        self._write()

    def close(self):
        try:
            self.selector.unregister(self.sock)
        except Exception as e:
            print(
                f"error: selector.unregister() exception for",
                f"{self.addr}: {repr(e)}",
            )

        try:
            self.sock.close()
        except OSError as e:
            print(
                f"error: socket.close() exception for",
                f"{self.addr}: {repr(e)}",
            )
        finally:
            # Delete reference to socket object for garbage collection
            self.sock = None

    def process_odheader(self):
        hdrlen = 10
        if len(self._recv_buffer) >= hdrlen:
            self._payload_len = struct.unpack('=i',self._recv_buffer[0:4])[0]
            self._reqid = struct.unpack('=i',self._recv_buffer[4:8])[0]
            self._subid = struct.unpack('=h',self._recv_buffer[8:hdrlen])[0]
            self._recv_buffer = self._recv_buffer[hdrlen:]

    def process_jsonheader(self):
        if len(self._recv_buffer) >= 4:
            (self._jsonheader_len,self.jsonheader,self._recv_buffer) = \
                self._json_decode(
                    self._recv_buffer, "utf-8"
            )
            for reqhdr in (
                "byteorder",
                "content-length",
                "content-type",
                "content-encoding",
            ):
                if reqhdr not in self.jsonheader:
                    raise ValueError(f'Missing required header "{reqhdr}".')

    def process_request(self):
        content_len = self.jsonheader["content-length"]
        if not len(self._recv_buffer) >= content_len:
            return
        data = self._recv_buffer[:content_len]
        self._recv_buffer = self._recv_buffer[content_len:]
        if self.jsonheader["content-type"] == "text/json":
            encoding = self.jsonheader["content-encoding"]
            (jsonsz,self.request,self._recv_buffer) = \
                                 self._json_decode(data, encoding)
        elif self.jsonheader["content-type"] == 'binary/array':
            shapes = self.jsonheader['array-shape']
            dtypes = self.jsonheader['content-encoding']
            self.request = self._array_decode(data,shapes,dtypes)
        else:
            # Binary or unknown content-type
            self.request = data
            print(
                f'received {self.jsonheader["content-type"]} request from',
                self.addr,
            )
        # Set selector to listen for write events, we're done reading.
        self._set_selector_events_mask("w")

    def create_response(self):
        if self.jsonheader["content-type"] == 'text/json':
            response = self._create_response_json_content()
        elif self.jsonheader["content-type"] == 'binary/array':
            (self,response) = self._create_response_array_content()
        else:
            # Binary or unknown content-type
            response = self._create_response_binary_content()
        message = self._create_message(**response)
        self.response_created = True
        self._send_buffer += message



#
# (C) dGB Beheer B.V.; (LICENSE) http://opendtect.org/OpendTect_license.txt
# AUTHOR   : Bert
# DATE     : Jan 2019
#
# Provides a test server for binary data transfer via stdin and stdout
#
# The client has to put:
# - a 4-byte int holding the number of values
# - a block of floats
# Then the script will return:
# - the number of returned vals (always 2)
# - a block of floats (the avg and std of the values passed)
# This stops when the input nrvals < 0
#

import sys
import struct
import numpy
import time
import platform

inpstrm = sys.stdin.buffer
outstrm = sys.stdout.buffer

if platform.system() == 'Windows':
  def dbg_pr( what, val ):
    pass
else:
  dbg_strm = open( "/tmp/dbg.txt", "w" )
  def dbg_pr( what, val ):
    dbg_strm.write( what + ": " + val + "\n" )
    dbg_strm.flush()

dbg_pr( "Started", "script" )

def put_to_output( what ):
  return outstrm.write( what )

def get_from_input( nr ):
  return inpstrm.read( nr )

def mk_int_bytes( ival ):
  return ival.to_bytes( 4, byteorder=sys.byteorder, signed=True )

def get_int_from_bytes( data_read ):
  return int.from_bytes( data_read, byteorder=sys.byteorder, signed=True )


outvals = numpy.zeros( 2, dtype=numpy.float32 )
time.sleep( 2 ) # simulate startup time
dbg_pr( "Started", "loop" )

while True:

  dbg_pr( "Try get", "4 bytes for nrvals" )
  try:
    data_read = get_from_input( 4 )
  except:
    dbg_pr( "Failed", "getting 4 bytes nrvals" )
    break

  nrvals = get_int_from_bytes( data_read )
  dbg_pr( "Result nrvals", str(nrvals) )
  if nrvals == 0:
    time.sleep( 0.01 )
    dbg_pr( "Got 0 vals, waited", "0.01 seconds" )
    continue
  if nrvals < 0:
    dbg_pr( "Stopped", "loop" )
    break

  dbg_pr( "Try get", str(nrvals) + " values" )
  try:
    inpdata = get_from_input( 4*nrvals )
  except:
    dbg_pr( "Exit", "except" )
    exit( 1 )
  dbg_pr( "Got", str(nrvals) + " values" )

  vals = struct.unpack( 'f'*nrvals, inpdata )
  dbg_pr( "Got values", str(vals) )
  time.sleep( 2 ) # simulate calculation time
  outvals[0] = numpy.mean(vals)
  outvals[1] = numpy.std(vals)
  dbg_pr( "Calculated values", str(outvals) )

  nrbyteswritten = put_to_output( mk_int_bytes(len(outvals)) )
  dbg_pr( "Put", "number of bytes, nrbyteswritten=" + str(nrbyteswritten) )
  nrbyteswritten = put_to_output( outvals.tobytes() )
  dbg_pr( "Put", "outvals, nrbyteswritten=" + str(nrbyteswritten) )

dbg_pr( "Exit", "normal" )
exit( 0 )



#
# (C) dGB Beheer B.V.; (LICENSE) http://opendtect.org/OpendTect_license.txt
# AUTHOR   : Wayne Mogg
# DATE     : June 2020
#
# dGB Keras machine learning models in UserModel format
#

from keras.models import *
from keras.layers import *
from keras.optimizers import Adam, RMSprop
from keras import backend as kb

from dgbpy.keras_classes import UserModel, DataPredType, OutputType, DimType
import dgbpy.dgbkeras as dgbkeras

def _to_tensor(x, dtype):
  from keras.optimizers import tf
  x = tf.convert_to_tensor(value=x)
  if x.dtype != dtype:
    x = tf.cast(x, dtype)
  return x

def root_mean_squared_error(y_true, y_pred):
  return kb.sqrt(kb.mean(kb.square(y_pred - y_true)))

def cross_entropy_balanced(y_true, y_pred):
  from keras.models import K
  from keras.optimizers import tf
  _epsilon = _to_tensor(K.epsilon(), y_pred.dtype.base_dtype)
  y_pred   = tf.clip_by_value(y_pred, _epsilon, 1 - _epsilon)
  y_pred   = tf.math.log(y_pred/ (1 - y_pred))

  y_true = tf.cast(y_true, tf.float32)
  count_neg = tf.reduce_sum(input_tensor=1. - y_true)
  count_pos = tf.reduce_sum(input_tensor=y_true)
  beta = count_neg / (count_neg + count_pos)
  pos_weight = beta / (1 - beta)
  cost = tf.nn.weighted_cross_entropy_with_logits(logits=y_pred, labels=y_true, pos_weight=pos_weight)
  cost = tf.reduce_mean(input_tensor=cost * (1 - beta))
  return tf.compat.v1.where(tf.equal(count_pos, 0.0), 0.0, cost)

def dGBUNet(model_shape, nroutputs, predtype):
    input_shape = model_shape
    data_format = 'channels_last'
      
    ndim = dgbkeras.getModelDims( model_shape, data_format )
    
    conv = pool = upsamp = None
    if ndim==3:
      conv = Conv3D
      pool = MaxPooling3D
      upsamp = UpSampling3D
    elif ndim==2:
      conv = Conv2D
      pool = MaxPooling2D
      upsamp = UpSampling2D
    elif ndim==1:
      conv = Conv1D
      pool = MaxPooling1D
      upsamp = UpSampling1D
      
    unet_smallsz = (2,64)
    unet_mediumsz = (16,512)
    unet_largesz = (32,512)

    unetnszs= unet_mediumsz
    poolsz1 = 2
    poolsz2 = 2
    poolsz3 = 2
    upscalesz = 2
    filtersz1 = unetnszs[0]
    filtersz2 = filtersz1 * poolsz2
    filtersz3 = filtersz2 * poolsz3
    filtersz4 = unetnszs[1]
    axis = -1

    params = dict(kernel_size=3, activation='relu', padding='same', data_format=data_format)

    inputs = Input(input_shape)
    conv1 = conv(filtersz1, **params)(inputs)
    conv1 = conv(filtersz1, **params)(conv1)
    
    pool1 = pool(pool_size=poolsz1, data_format=data_format)(conv1)
    
    conv2 = conv(filtersz2, **params)(pool1)
    conv2 = conv(filtersz2, **params)(conv2)
    pool2 = pool(pool_size=poolsz2, data_format=data_format)(conv2)
    
    conv3 = conv(filtersz3, **params)(pool2)
    conv3 = conv(filtersz3, **params)(conv3)
    pool3 = pool(pool_size=poolsz3, data_format=data_format)(conv3)
    
    conv4 = conv(filtersz4, **params)(pool3)
    conv4 = conv(filtersz4, **params)(conv4)
    
    up5 = concatenate([upsamp(size=upscalesz,data_format=data_format)(conv4), conv3], axis=axis)
    conv5 = conv(filtersz3, **params)(up5)
    conv5 = conv(filtersz3, **params)(conv5)
    
    up6 = concatenate([upsamp(size=poolsz2,data_format=data_format)(conv5), conv2], axis=axis)
    conv6 = conv(filtersz2, **params)(up6)
    conv6 = conv(filtersz2, **params)(conv6)
    
    up7 = concatenate([upsamp(size=poolsz1,data_format=data_format)(conv6), conv1], axis=axis)
    conv7 = conv(filtersz1, **params)(up7)
    conv7 = conv(filtersz1, **params)(conv7)
    
    if isinstance(predtype, DataPredType) and predtype==DataPredType.Continuous:
      nrout = nroutputs
      activation = 'linear'
    else:
      if nroutputs == 2:
        nrout = 1
        activation = 'sigmoid'
      else:
        nrout = nroutputs
        activation = 'softmax'
    conv8 = conv(nrout, 1, activation=activation, data_format=data_format)(conv7)
      
    model = Model(inputs=[inputs], outputs=[conv8])
    return model

class dGB_UnetSeg(UserModel):
  uiname = 'dGB UNet Segmentation'
  uidescription = 'dGBs Unet image segmentation'
  predtype = DataPredType.Classification
  outtype = OutputType.Image
  dimtype = DimType.Any
  
  def _make_model(self, model_shape, nroutputs, learnrate):
    model = dGBUNet(model_shape, nroutputs, self.predtype)
    if nroutputs<=2:
      loss = cross_entropy_balanced
    else:
      loss = 'categorical_crossentropy'

    model.compile(optimizer = Adam(lr = learnrate), loss = loss, metrics = ['accuracy'])

    return model
  
class dGB_UnetReg(UserModel):
  uiname = 'dGB UNet Regression'
  uidescription = 'dGBs Unet image regression'
  predtype = DataPredType.Continuous
  outtype = OutputType.Image
  dimtype = DimType.Any
  
  def _make_model(self, model_shape, nroutputs, learnrate):
    model = dGBUNet(model_shape, nroutputs, self.predtype)
    model.compile(optimizer = Adam(lr = learnrate), loss = 'mse', metrics = ['mae'])
    return model
  
def dGBLeNet(model_shape, nroutputs, predtype):
    
  input_shape = model_shape
  data_format = 'channels_last'
  bnaxis = -1
  if not dgbkeras.need_channels_last():
  #Prefer channels_first data_format for efficiency if supported
  # CPU training requires channels_last format due to TensorFlow bug
      data_format = 'channels_first'
      input_shape = (model_shape[-1], *model_shape[0:-1])
      bnaxis = 1
      
  ndim = dgbkeras.getModelDims(model_shape, data_format)

  conv = None
  if ndim==3:
    conv = Conv3D
  elif ndim==2:
    conv = Conv2D
  elif ndim==1 or ndim==0:
    conv = Conv1D
    
  filtersz = 50
  densesz = 10  
  kernel_sz1 = 5
  kernel_sz2 = 3
  stride_sz1 = 4
  stride_sz2 = 2
  dropout = 0.2

  inputs = Input(input_shape)
  conv1 = conv(filtersz, kernel_sz1, strides=stride_sz1, padding='same', data_format=data_format)(inputs)
  conv1 = BatchNormalization(axis=bnaxis)(conv1)
  conv1 = Activation('relu')(conv1)
  conv2 = conv(filtersz, kernel_sz2, strides=stride_sz2, padding='same', data_format=data_format)(conv1)
  conv2 = Dropout(dropout)(conv2)
  conv2 = BatchNormalization(axis=bnaxis)(conv2)
  conv2 = Activation('relu')(conv2)
  conv3 = conv(filtersz, kernel_sz2, strides=stride_sz2, padding='same', data_format=data_format)(conv2)
  conv3 = Dropout(dropout)(conv3)
  conv3 = BatchNormalization(axis=bnaxis)(conv3)
  conv3 = Activation('relu')(conv3)
  conv4 = conv(filtersz, kernel_sz2, strides=stride_sz2, padding='same', data_format=data_format)(conv3)
  conv4 = Dropout(dropout)(conv4)
  conv4 = BatchNormalization(axis=bnaxis)(conv4)
  conv4 = Activation('relu')(conv4)
  conv5 = conv(filtersz, kernel_sz2, strides=stride_sz2, padding='same', data_format=data_format)(conv4)
  dense1 = Flatten()(conv5)
  dense1 = Dense(filtersz)(dense1)
  dense1 = BatchNormalization(axis=bnaxis)(dense1)
  dense1 = Activation('relu')(dense1)
  dense2 = Dense(densesz)(dense1)
  dense2 = BatchNormalization(axis=bnaxis)(dense2)
  dense2 = Activation('relu')(dense2)
  dense3 = None
  if isinstance(predtype, DataPredType) and predtype==DataPredType.Continuous:
    dense3 = Dense(nroutputs, activation='linear')(dense2)
  else:
    dense3 = Dense(nroutputs)(dense2)
    dense3 = BatchNormalization(axis=bnaxis)(dense3)
    dense3 = Activation('softmax')(dense3)
    
  model = Model(inputs=[inputs], outputs=[dense3])
  return model
        
class dGB_LeNet_Classifier(UserModel):
  uiname = 'dGB LeNet classifier'
  uidescription = 'dGBs LeNet classifier Keras model in UserModel form'
  predtype = DataPredType.Classification
  outtype = OutputType.Pixel
  dimtype = DimType.Any
  
  def _make_model(self, input_shape, nroutputs, learnrate):
    model = dGBLeNet(input_shape, nroutputs, self.predtype)
    
    loss = 'binary_crossentropy'
    if nroutputs>2:
      loss = 'categorical_crossentropy'
    
    model.compile(optimizer=Adam(lr=learnrate), loss=loss, metrics=['accuracy'])
    
    return model

class dGB_LeNet_Regressor(UserModel):
  uiname = 'dGB LeNet regressor'
  uidescription = 'dGBs LeNet regressor Keras model in UserModel form'
  predtype = DataPredType.Continuous
  outtype = OutputType.Pixel
  dimtype = DimType.Any
  
  def _make_model(self, input_shape, nroutputs, learnrate):
    model = dGBLeNet(input_shape, nroutputs, self.predtype)
      
    model.compile(optimizer=Adam(lr=learnrate), loss='mse', metrics=['mae'])
    
    return model




#
# (C) dGB Beheer B.V.; (LICENSE) http://opendtect.org/OpendTect_license.txt
# AUTHOR   : A. Huck
# DATE     : Nov 2018
#
# Pre-loads (part of) an HDF5 seismic dataset created by OpendTect, with sampling as:
# sampling = {
#  'Inline': range(start,stop,step)
#  'Crossline': range(start,stop,step)
#  'Z': range(start,stop,step)
# }
# Mind that in python stop >= start+step, otherwise the range is empty!
# The entire sampling can be retrieve with odhdf5.getAttribInfo( filenm )['range']
#

import numpy as np
import h5py
import odpy.hdf5 as odhdf5
import odpy.ranges as ranges

def getAxes( sampling, dataarr ):
  flatx1 = len(dataarr) < 2
  flatx2 = len(dataarr[0]) < 2
  flatx3 = len(dataarr[0][0]) < 2
  lines = ranges.getLineObj( sampling )
  traces = ranges.getTraceObj( sampling )
  zsamp = ranges.getZObj( sampling )
  x3axis = None
  if flatx1:
    x1axis = traces
    x2axis = zsamp
  elif flatx2:
    x1axis = lines
    x2axis = zsamp
  elif flatx3:
    x1axis = lines 
    x2axis = traces
  else:
    x1axis = lines 
    x2axis = traces
    x3axis = zsamp

  x1label = list(sampling.keys())[list(sampling.values()).index(x1axis)]
  x2label = list(sampling.keys())[list(sampling.values()).index(x2axis)]
  x3label = None
  if x3axis != None:
    x3label = list(sampling.keys())[list(sampling.values()).index(x3axis)]

  return {
    'range': [ranges.arrayFromRange(x1axis),
              ranges.arrayFromRange(x2axis),
              ranges.arrayFromRange(x3axis)],
    'labels': [x1label,x2label,x3label]
  }

def preLoad( filenm, attribute=None, sampling=None, fillval=None ):
  attribinfo = odhdf5.getAttribInfo( filenm )
  samplinginfo = attribinfo['range']
  datatkzs = ranges.getAxesAsRanges( samplinginfo )
  if attribute==None:
    attribinfo = odhdf5.getAttribInfo( filenm )
    compnms = attribinfo['attributes']
    if isinstance( compnms, list ):
      if len(compnms)<1:
        return []
      attribute = compnms[0]
    else:
      attribute = compnms

  if sampling==None:
    sampling = datatkzs
  if fillval==None:
    fillval = np.nan

  blocksinfo = attribinfo['block']
  blocksz = blocksinfo['size']
  dim1blocksz = blocksz[0]
  dim2blocksz = blocksz[1]
  blocksrg = blocksinfo['range']
  nrblockx = blocksrg['Inline'][1]+1
  nrblocky = blocksrg['Crossline'][1]+1

  h5file = odhdf5.openFile( filenm, 'r' )
  datagroup = h5file[attribute]
  if len(datagroup) < 1:
    print( "Empty dataset found" )
    return []

  alllines = ranges.getLineObj( datatkzs )
  alltraces = ranges.getTraceObj( datatkzs )
  allzsamp = ranges.getZObj( datatkzs )
  lines = ranges.getLineObj( sampling )
  traces = ranges.getTraceObj( sampling )
  zsamp = ranges.getZObj( sampling )
  dim3blocksz = len(allzsamp)
  nrz = len(zsamp)
  blkzrg = [ allzsamp[0], allzsamp[-1] ]
  zrg = [ max(blkzrg[0],zsamp.start), min(blkzrg[1],zsamp.stop-zsamp.step) ]
  zrgin = [ allzsamp.index(zrg[0]) % dim3blocksz, allzsamp.index(zrg[1]) % dim3blocksz+1 ]
  zrgout = [ zsamp.index(zrg[0]), zsamp.index(zrg[1])+1 ]

  datagroupset = list( datagroup.items() )
  datasetsamptype = datagroupset[0][1].dtype
  sliceout = np.full((len(lines),len(traces),nrz),fillval,dtype=datasetsamptype)
  for blkidx in range(nrblockx):
    for blkidy in range(nrblocky):
      subcubenm = str(blkidx)+"."+str(blkidy)
      subcube = datagroup.get( subcubenm )
      if subcube == None:
        continue
      dim1shift = 0
      dim2shift = 0
      if len(subcube.attrs['Loc00']) > 0:
        locorig = odhdf5.getIInterval( subcube, "Loc00" )
        dim1shift = locorig[0]
        dim2shift = locorig[1]
      firstx1 = blkidx * dim1blocksz + dim1shift
      lastx1 = firstx1 + subcube.shape[0] - 1
      blklinerg = [ alllines[firstx1], alllines[lastx1] ]
      if lines.stop-lines.step < blklinerg[0] or lines.start > blklinerg[1]:
        continue
      firstx2 = blkidy * dim2blocksz + dim2shift
      lastx2 = firstx2 + subcube.shape[1] - 1
      blktrcrg = [ alltraces[firstx2], alltraces[lastx2] ]
      if traces.stop-traces.step < blktrcrg[0] or traces.start > blktrcrg[1]:
        continue
      linerg = [ max(blklinerg[0],lines.start), min(blklinerg[1],lines.stop-lines.step) ]
      trcrg  = [ max(blktrcrg[0],traces.start), min(blktrcrg[1],traces.stop-traces.step) ]
      linergin = [ alllines.index(linerg[0]) % dim1blocksz-dim1shift,
                   alllines.index(linerg[1]) % dim1blocksz+1-dim1shift ]
      trcrgin = [ alltraces.index(trcrg[0]) % dim2blocksz-dim2shift,
                  alltraces.index(trcrg[1]) % dim2blocksz+1-dim2shift ]
      linergout = [ lines.index(linerg[0]), lines.index(linerg[1])+1 ]
      trcrgout = [traces.index(trcrg[0]), traces.index(trcrg[1])+1 ]
      sliceout[linergout[0]:linergout[1],trcrgout[0]:trcrgout[1],zrgout[0]:zrgout[1]] = subcube[linergin[0]:linergin[1],trcrgin[0]:trcrgin[1],zrgin[0]:zrgin[1]]
  h5file.close()

  axes = getAxes( sampling, sliceout )
  sliceout = np.squeeze(sliceout)

  return {
    'data': sliceout,
    'axes': axes
  }



# __________________________________________________________________________
#
# (C) dGB Beheer B.V.; (LICENSE) http://opendtect.org/OpendTect_license.txt
# Author:        A. Huck
# Date:          Apr 2019
#
# _________________________________________________________________________

from functools import partial

import numpy as np

from bokeh.layouts import column
from bokeh.models.widgets import CheckboxGroup, Div, Select, Slider

from odpy.common import log_msg
from dgbpy.dgbkeras import *  # noqa ignore=F405
from dgbpy import uibokeh

info = None


def getPlatformNm(full=False):
    if full:
        return platform
    return getMLPlatform()


def getSizeStr(sizeinbytes):
    ret = 'Size: '
    try:
        import humanfriendly
        ret += humanfriendly.format_size(sizeinbytes)
    except Exception as e:
        ret += str(int(sizeinbytes)) + ' bytes'
    return ret


def getUiModelTypes(learntype, classification, ndim):
    ret = ()
    for model in getModelsByType(learntype, classification, ndim):
        ret += ((model,),)

    return dgbkeys.getNames(ret)


def getUiPars(uipars=None):
    dict = keras_dict
    learntype = info[dgbkeys.learntypedictstr]
    if isinstance(info[dgbkeys.inpshapedictstr], int):
        ndim = 1
    else:
        ndim = len(info[dgbkeys.inpshapedictstr])
    modeltypes = getUiModelTypes(learntype, info[dgbkeys.classdictstr], ndim)
    if len(modeltypes) == 0:
        divfld = Div(text="""No Keras models found for this workflow.""")
        parsgrp = column(divfld)
        return {'grp': parsgrp,
                'uiobjects': {
                    'divfld': divfld
                }
                }
    defmodel = modeltypes[0]
    defbatchsz = keras_dict['batch']
    estimatedsz = info[dgbkeys.estimatedsizedictstr]
    if kc.UserModel.isImg2Img(defmodel):
        defbatchsz = 4
    uiobjs = {}
    if not uipars:
        uiobjs = {
            'modeltypfld': Select(title='Type', options=modeltypes),
            'batchfld': Select(title='Batch Size', options=cudacores),
            'epochfld': Slider(start=1, end=1000, title='Epochs'),
            'patiencefld': Slider(start=1, end=100, title='Patience'),
            'lrfld': Slider(start=-10, end=-1, step=1, title='Initial Learning Rate (1e)'),
            'edfld': Slider(start=1, end=100, title='Epoch drop (%)', step=0.1),
            'sizefld': None,
            'dodecimatefld': CheckboxGroup(labels=['Decimate input']),
            'chunkfld': Slider(start=1, end=100, title='Number of Chunks'),
            'rundevicefld': CheckboxGroup(labels=['Train on GPU'], visible=can_use_gpu())
        }
        if estimatedsz:
            uiobjs['sizefld'] = Div(text=getSizeStr(estimatedsz))
        uiobjs['dodecimatefld'].on_click(
            partial(decimateCB, chunkfld=uiobjs['chunkfld'], sizefld=uiobjs['sizefld']))
        try:
            uiobjs['chunkfld'].value_throttled = uiobjs['chunkfld'].value
            uiobjs['chunkfld'].on_change(
                'value_throttled', partial(chunkfldCB, uiobjs['sizefld']))
        except AttributeError:
            log_msg('[WARNING] Bokeh version too old, consider updating it.')
            pass
        parsgrp = column(*list(uiobjs.values()))
        uipars = {'grp': parsgrp, 'uiobjects': uiobjs}
    else:
        uiobjs = uipars['uiobjects']

    uiobjs['modeltypfld'].value = defmodel
    uiobjs['batchfld'].value = str(defbatchsz)
    uiobjs['epochfld'].value = dict['epoch']
    uiobjs['patiencefld'].value = dict['patience']
    uiobjs['lrfld'].value = np.log10(dict['learnrate'])
    uiobjs['edfld'].value = 100*dict['epochdrop']/uiobjs['epochfld'].value
    if estimatedsz:
        uiobjs['sizefld'].text = getSizeStr(estimatedsz)
    uiobjs['dodecimatefld'].active = []
    uiobjs['chunkfld'].value = dict['nbchunk']
    uiobjs['rundevicefld'].active = [0]
    decimateCB(uiobjs['dodecimatefld'].active,
               uiobjs['chunkfld'], uiobjs['sizefld'])
    return uipars


def chunkfldCB(sizefld, attr, old, new):
    if sizefld == None:
        return
    sizefld.text = getSizeStr(info[dgbkeys.estimatedsizedictstr]/new)


def getUiParams(keraspars):
    kerasgrp = keraspars['uiobjects']
    nrepochs = kerasgrp['epochfld'].value
    epochdroprate = kerasgrp['edfld'].value / 100
    epochdrop = int(nrepochs*epochdroprate)
    if epochdrop < 1:
        epochdrop = 1
    runoncpu = not kerasgrp['rundevicefld'].visible or \
        not isSelected(kerasgrp['rundevicefld'])
    return getParams(dodec=isSelected(kerasgrp['dodecimatefld']),
                     nbchunk=kerasgrp['chunkfld'].value,
                     epochs=kerasgrp['epochfld'].value,
                     batch=int(kerasgrp['batchfld'].value),
                     patience=kerasgrp['patiencefld'].value,
                     learnrate=10 ** kerasgrp['lrfld'].value,
                     epochdrop=epochdrop,
                     nntype=kerasgrp['modeltypfld'].value,
                     prefercpu=runoncpu)


def isSelected(fldwidget, index=0):
    return uibokeh.integerListContains(fldwidget.active, index)


def decimateCB(widgetactivelist, chunkfld, sizefld):
    decimate = uibokeh.integerListContains(widgetactivelist, 0)
    chunkfld.visible = decimate
    if sizefld == None:
        return
    size = info[dgbkeys.estimatedsizedictstr]
    if decimate:
        size /= chunkfld.value
    sizefld.text = getSizeStr(size)



#__________________________________________________________________________
#
# (C) dGB Beheer B.V.; (LICENSE) http://opendtect.org/OpendTect_license.txt
# Author:        A. Huck
# Date:          Jan 2019
#
# _________________________________________________________________________
# Standard keys used by the dgb machine learning python modules
#


# Dictionary keys (lower case, mostly alphabetic only):

classdictstr = 'classification'
classesdictstr = 'classes'
classnmdictstr = 'classnm'
collectdictstr = 'collection'
componentdictstr = 'component'
confdictstr = 'confidence'
datasetdictstr = 'datasets'
dbkeydictstr = 'dbkey'
decimkeystr = 'decimation'
dtypeconf = 'confdtype'
dtypepred = 'preddtype'
dtypeprob = 'probdtype'
estimatedsizedictstr = 'estimatedsize'
exampledictstr = 'examples'
filedictstr = 'filenm'
geomiddictstr = 'geomid'
iddictstr = 'id'
infodictstr = 'info'
inpshapedictstr = 'inp_shape'
inputdictstr = 'input'
interpoldictstr = 'interpolated'
learntypedictstr = 'learntype'
locationdictstr = 'location'
logdictstr = 'log'
namedictstr = 'name'
nroutdictstr = 'nroutputs'
pathdictstr = 'path'
plfdictstr = 'platform'
preddictstr = 'prediction'
probadictstr = 'probabilities'
rangedictstr = 'range'
scaledictstr = 'scale'
outshapedictstr = 'out_shape'
surveydictstr = 'survey'
targetdictstr = 'target'
traindictstr = 'train'
trainseldicstr = 'training_selection'
validdictstr = 'validation'
versiondictstr = 'version'
xdatadictstr = 'x_data'
xtraindictstr = 'x_train'
xvaliddictstr = 'x_validate'
ydatadictstr = 'y_data'
ytraindictstr = 'y_train'
yvaliddictstr = 'y_validate'
ypreddictstr = 'y_pred'
zstepdictstr = 'zstep'


# Value keys

classdatavalstr = 'Classification Data'
classesvalstr = 'Classes'
classificationvalstr = 'Classification'
continuousvalstr = 'Continuous Data'
confvalstr = 'Confidence'
contentvalstr = 'Content'
inpshapestr = 'Input.Shape'
kerasplfnm = 'keras'
loglogtypestr = 'Log-Log Prediction'
mlsoftkey = 'OpendTect-ML-Software'
modelnm = 'new model'
numpyvalstr = 'numpy'
outshapestr = 'Output.Shape'
scikitplfnm = 'scikit'
seisclasstypestr = 'Seismic Classification'
seisimgtoimgtypestr = 'Seismic Image Transformation'
seisproptypestr = 'Property Prediction from Seismics'
typestr = 'Type'
valuestr = 'Value'
versionstr = 'Version'

def getNames( lst, uinames=True ):
  idx = -1
  if not uinames:
    idx = 0
  ret = list()
  for itm in lst:
    ret.append( itm[idx] )
  return ret

def getNameFromUiName( lst, itmnm ):
  for lstitms in lst:
    if lstitms[1] == itmnm:
      return lstitms[0]
  return None

def getNameFromList( lst, itmnm, uiname ):
  for lstitms in lst:
    if lstitms[0] != itmnm and lstitms[1] != itmnm:
      continue
    if uiname:
      return lstitms[1]
    else:
      return lstitms[0]
  return None



# __________________________________________________________________________
#
# (C) dGB Beheer B.V.; (LICENSE) http://opendtect.org/OpendTect_license.txt
# Author:        A. Huck
# Date:          Apr 2019
#
# _________________________________________________________________________

from functools import partial

from bokeh.core import enums
from bokeh.layouts import row, column
from bokeh.models import Spacer
from bokeh.models.widgets import Button, Select, Slider

from odpy.common import log_msg
from dgbpy.dgbscikit import * # noqa ignore=F405
from dgbpy import uibokeh

but_width = uibokeh.but_width
but_height = uibokeh.but_height
but_spacer = uibokeh.but_spacer
defaultbut = enums.ButtonType.default


def getPlatformNm(full=False):
    if full:
        return platform
    return getMLPlatform()


def getLinearGrp(uipars=None):
    uiobjs = {}
    if not uipars:
        uiobjs = {'lineartyp': Select(
            title='Model', options=getUiLinearTypes()), }
        uipars = {'uiobjects': uiobjs, 'name': regmltypes[0][1], }
    else:
        uiobjs = uipars['uiobjects']

    uiobjs['lineartyp'].value = 'Ordinary Least Squares'
    return uipars


def getLogGrp(uipars=None):
    uiobjs = {}
    if not uipars:
        uiobjs = {'logtyp': Select(title='Model', options=getUiLogTypes()),
                  'solvertyp': Select(title='Solver', options=getUiSolverTypes())
                  }
        uipars = {'uiobjects': uiobjs, 'name': classmltypes[0][1], }
    else:
        uiobjs = uipars['uiobjects']

    uiobjs['logtyp'].value = 'Logistic Regression Classifier'
    uiobjs['solvertyp'].value = getDefaultSolver()
    return uipars


def getEnsembleGrp(uipars=None):
    uiobjs = {}
    rfgrp = {}
    xgdtgrp = None
    xgrfgrp = None
    if uipars:
        uiobjs = uipars['uiobjects']
        if hasXGBoost():
            xgdtgrp = {'uiobjects': {'depparfldxgdt': uiobjs['depparfldxgdt'],
                                     'estparfldxgdt': uiobjs['estparfldxgdt'],
                                     'lrparfldxgdt': uiobjs['lrparfldxgdt'], }, }
            xgrfgrp = {'uiobjects': {'depparfldxgrf': uiobjs['depparfldxgrf'],
                                     'estparfldxgrf': uiobjs['estparfldxgrf'],
                                     'lrparfldxgrf': uiobjs['lrparfldxgrf'], }, }
        rfgrp = {'uiobjects': {'depparfldrf': uiobjs['depparfldrf'],
                               'estparfldrf': uiobjs['estparfldrf'], }, }
        gbgrp = {'uiobjects': {'depparfldgb': uiobjs['depparfldgb'],
                               'estparfldgb': uiobjs['estparfldgb'],
                               'lrparfldgb': uiobjs['lrparfldgb'], }, }
        adagrp = {'uiobjects': {'estparfldada': uiobjs['estparfldada'],
                                'lrparfldada': uiobjs['lrparfldada'], }, }
        if hasXGBoost():
            getXGDTGrp(xgdtgrp)
            getXGRFGrp(xgrfgrp)
        getRFGrp(rfgrp)
        getGBGrp(gbgrp)
        getAdaGrp(adagrp)
    else:
        if hasXGBoost():
            xgdtgrp = getXGDTGrp()
            xgrfgrp = getXGRFGrp()
        rfgrp = getRFGrp()
        gbgrp = getGBGrp()
        adagrp = getAdaGrp()
        uiobjs = {'ensembletyp': Select(title='Model',
                                        options=getUiEnsembleTypes())}
        if hasXGBoost():
            xggrpuiobjs = xgdtgrp['uiobjects']
            uiobjs.update({
                'depparfldxgdt': xggrpuiobjs['depparfldxgdt'],
                'estparfldxgdt': xggrpuiobjs['estparfldxgdt'],
                'lrparfldxgdt': xggrpuiobjs['lrparfldxgdt'],
            })
            xggrpuiobjs = xgrfgrp['uiobjects']
            uiobjs.update({
                'depparfldxgrf': xggrpuiobjs['depparfldxgrf'],
                'estparfldxgrf': xggrpuiobjs['estparfldxgrf'],
                'lrparfldxgrf': xggrpuiobjs['lrparfldxgrf'],
            })
        uiobjs.update({
            'depparfldrf': rfgrp['uiobjects']['depparfldrf'],
            'estparfldrf': rfgrp['uiobjects']['estparfldrf'],
            'depparfldgb': gbgrp['uiobjects']['depparfldgb'],
            'estparfldgb': gbgrp['uiobjects']['estparfldgb'],
            'lrparfldgb': gbgrp['uiobjects']['lrparfldgb'],
            'estparfldada': adagrp['uiobjects']['estparfldada'],
            'lrparfldada': adagrp['uiobjects']['lrparfldada'],
        })
        if hasXGBoost():
            ensemblegrp = (xgdtgrp, xgrfgrp)
        ensemblegrp += (rfgrp, gbgrp, adagrp,)
        uiobjs['ensembletyp'].on_change('value', partial(
            ensembleChgCB, cb=uiobjs['ensembletyp'], ensemblegrp=ensemblegrp))
        uipars = {'uiobjects': uiobjs, 'name': regmltypes[1][1], }

    if hasXGBoost():
        uiobjs['ensembletyp'].value = 'XGBoost: (Decision Tree)'
    else:
        uiobjs['ensembletyp'].value = 'Random Forests'
    return uipars


def getNNGrp(uipars=None):
    dict = scikit_dict
    uiobjs = {}
    if not uipars:
        uiobjs = {'nntyp': Select(title='Model', options=getUiNNTypes()),
                  'itrparfld': Slider(start=10, end=1000, title='Max Iteration'),
                  'lay1parfld': Slider(start=1, end=200, title='Layer 1'),
                  'lay2parfld': Slider(start=1, end=50, title='Layer 2'),
                  'lay3parfld': Slider(start=1, end=10, title='Layer 3'),
                  'lay4parfld': Slider(start=1, end=5, title='Layer 4'),
                  'lay5parfld': Slider(start=1, end=3, title='Layer 5'),
                  'lrparfld': Slider(start=1, end=100, title='Initial Learning Rate ' + '('+u'\u2030'+')'),
                  'addbutton': Button(label='Add', button_type=defaultbut, width=but_width, height=but_height),
                  'lessbutton': Button(label='Less', button_type=defaultbut, width=but_width, height=but_height),
                  }
        uiobjs['buttonparfld'] = row(uiobjs['addbutton'], Spacer(
            width=5), uiobjs['lessbutton'], sizing_mode='stretch_width')
        nb = Slider(start=1, end=5, value=3)
        layergrp = [nb, uiobjs['lay1parfld'], uiobjs['lay2parfld'], uiobjs['lay3parfld'],
                    uiobjs['lay4parfld'], uiobjs['lay5parfld'], uiobjs['addbutton'], uiobjs['lessbutton']]
        try:
            uiobjs['lay1parfld'].value_throttled = uiobjs['lay1parfld'].value
            uiobjs['lay2parfld'].value_throttled = uiobjs['lay2parfld'].value
            uiobjs['lay3parfld'].value_throttled = uiobjs['lay3parfld'].value
            uiobjs['lay4parfld'].value_throttled = uiobjs['lay4parfld'].value
            uiobjs['lay1parfld'].on_change(
                'value_throttled', partial(layer1ChgCB, layergrp))
            uiobjs['lay2parfld'].on_change(
                'value_throttled', partial(layer2ChgCB, layergrp))
            uiobjs['lay3parfld'].on_change(
                'value_throttled', partial(layer3ChgCB, layergrp))
            uiobjs['lay4parfld'].on_change(
                'value_throttled', partial(layer4ChgCB, layergrp))
        except AttributeError:
            log_msg('[WARNING] Bokeh version too old, consider updating it.')
            pass
        uiobjs['addbutton'].on_click(
            partial(buttonChgCB, uiobjs['addbutton'], layergrp))
        uiobjs['lessbutton'].on_click(
            partial(buttonChgCB, uiobjs['lessbutton'], layergrp))
        uipars = {'uiobjects': uiobjs,
                  'name': regmltypes[2][1],
                  'nb': nb
                  }
    else:
        uiobjs = uipars['uiobjects']

    uiobjs['nntyp'].value = 'Multi-Layer Perceptron'
    uiobjs['itrparfld'].value = dict['nnpars']['maxitr']
    uiobjs['lay1parfld'].value = 50
    uiobjs['lay2parfld'].value = dict['nnpars']['lay2']
    uiobjs['lay3parfld'].value = dict['nnpars']['lay3']
    uiobjs['lay4parfld'].value = dict['nnpars']['lay4']
    uiobjs['lay5parfld'].value = dict['nnpars']['lay5']
    uiobjs['lrparfld'].value = dict['nnpars']['lr']*1000
    uipars['nb'].value = 3
    return uipars


def getSVMGrp(uipars=None):
    dict = scikit_dict
    uiobjs = {}
    defkernelstr = getDefaultNNKernel()
    if not uipars:
        uiobjs = {'svmtyp': Select(title='Model', options=getUiSVMTypes()),
                  'kernel': Select(title='Kernel', options=getUiNNKernelTypes()),
                  'degree': Slider(start=1, end=5, step=1, title='Degree'),
                  }
        uiobjs['kernel'].on_change('value', partial(
            kernelChgCB, deg=uiobjs['degree']))
        kernelChgCB('value', defkernelstr, defkernelstr, uiobjs['degree'])
        uipars = {'uiobjects': uiobjs, 'name': regmltypes[3][1], }
    else:
        uiobjs = uipars['uiobjects']

    uiobjs['svmtyp'].value = 'Support Vector Machine'
    uiobjs['kernel'].value = defkernelstr
    uiobjs['degree'].value = dict['svmpars']['degree']
    return uipars

# deg visible only when polynomial kernel function


def kernelChgCB(attrnm, old, new, deg):
    if new == 'Polynomial':
        deg.visible = True
    else:
        deg.visible = False


def layer1ChgCB(layergrp, attr, old, new):
    layergrp[2].end = new
    if new <= layergrp[2].value:
        layergrp[2].value = new
        layer2ChgCB(layergrp, attr, layergrp[2].value, new)


def layer2ChgCB(layergrp, attr, old, new):
    layergrp[3].end = new
    if new <= layergrp[3].value:
        layergrp[3].value = new
        layer3ChgCB(layergrp, attr, layergrp[2].value, new)


def layer3ChgCB(layergrp, attr, old, new):
    layergrp[4].end = new
    if new <= layergrp[4].value:
        layergrp[4].value = new
        layer4ChgCB(layergrp, attr, layergrp[2].value, new)


def layer4ChgCB(layergrp, attr, old, new):
    layergrp[5].end = new
    if new <= layergrp[5].value:
        layergrp[5].value = new


def buttonChgCB(addbutton, layergrp):
    log_msg('Working')
    nb = layergrp[0].value  # number of layers
    if addbutton.label == 'Add':
        layergrp[nb+1].visible = True
        if nb == 4:
            layergrp[6].visible = False  # Hide add button
        elif nb == 1:
            layergrp[7].visible = True  # Show less button
        layergrp[0].value += 1
    elif addbutton.label == 'Less':
        layergrp[nb].visible = False
        if nb == 5:
            layergrp[6].visible = True  # show add button
        elif nb == 2:
            layergrp[7].visible = False  # Hide less button
        layergrp[0].value -= 1
    else:
        return None


def modelChgCB(attrnm, old, new, cb, modelsgrp):
    ret = uibokeh.getAllUiFlds(modelsgrp)
    for uifld in ret:
        uifld.visible = False
    ret = uibokeh.getGroup(new, cb.options, modelsgrp, 'uiobjects')
    # set default visibility
    if new == 'Ensemble':
        ret['ensembletyp'].visible = True
        ret['estparfldxgdt'].visible = True
        ret['depparfldxgdt'].visible = True
        ret['lrparfldxgdt'].visible = True
    elif new == 'SVM':
        ret['svmtyp'].visible = True
        ret['kernel'].visible = True
    else:
        for uifld in ret:
            ret[uifld].visible = True
        if new == 'Neural Network':
            ret['lay4parfld'].visible = False
            ret['lay5parfld'].visible = False


def getXGDTGrp(uipars=None):
    if not hasXGBoost():
        return None
    dict = scikit_dict['ensemblepars']['xgdt']
    uiobjs = {}
    if not uipars:
        uiobjs = {'estparfldxgdt': Slider(start=10, end=1000, step=10, title='Estimators'),
                  'depparfldxgdt': Slider(start=10, end=200, step=10, title='Max Depth'),
                  'lrparfldxgdt': Slider(start=0.1, end=10, step=0.1, title='Learning Rate'), }
        uipars = {'uiobjects': uiobjs, }
    else:
        uiobjs = uipars['uiobjects']

    uiobjs['estparfldxgdt'].value = dict['est']
    uiobjs['depparfldxgdt'].value = dict['maxdep']
    uiobjs['lrparfldxgdt'].value = dict['lr']
    return uipars


def getXGRFGrp(uipars=None):
    if not hasXGBoost():
        return None
    dict = scikit_dict['ensemblepars']['xgrf']
    uiobjs = {}
    if not uipars:
        uiobjs = {'estparfldxgrf': Slider(start=10, end=1000, step=10, title='Estimators'),
                  'depparfldxgrf': Slider(start=10, end=200, step=10, title='Max Depth'),
                  'lrparfldxgrf': Slider(start=0.1, end=10, step=0.1, title='Learning Rate'), }
        uipars = {'uiobjects': uiobjs, }
    else:
        uiobjs = uipars['uiobjects']

    uiobjs['estparfldxgrf'].value = dict['est']
    uiobjs['depparfldxgrf'].value = dict['maxdep']
    uiobjs['lrparfldxgrf'].value = dict['lr']
    return uipars


def getRFGrp(uipars=None):
    dict = scikit_dict['ensemblepars']['rf']
    uiobjs = {}
    if not uipars:
        uiobjs = {'estparfldrf': Slider(start=10, end=1000, step=10, title='Estimators'),
                  'depparfldrf': Slider(start=10, end=200, step=10, title='Max Depth'), }
        uipars = {'uiobjects': uiobjs}
    else:
        uiobjs = uipars['uiobjects']

    uiobjs['estparfldrf'].value = dict['est']
    uiobjs['depparfldrf'].value = dict['maxdep']
    return uipars


def getGBGrp(uipars=None):
    dict = scikit_dict['ensemblepars']['gb']
    uiobjs = {}
    if not uipars:
        uiobjs = {'estparfldgb': Slider(start=10, end=500, step=10, title='Estimators'),
                  'depparfldgb': Slider(start=1, end=100, step=1, title='Max Depth'),
                  'lrparfldgb': Slider(start=0.1, end=10, step=0.1, title='Learning Rate'), }
        uipars = {'uiobjects': uiobjs, }
    else:
        uiobjs = uipars['uiobjects']

    uiobjs['estparfldgb'].value = dict['est']
    uiobjs['depparfldgb'].value = dict['maxdep']
    uiobjs['lrparfldgb'].value = dict['lr']
    return uipars


def getAdaGrp(uipars=None):
    dict = scikit_dict['ensemblepars']['ada']
    uiobjs = {}
    if not uipars:
        uiobjs = {'estparfldada': Slider(start=10, end=500, step=10, title='Estimators'),
                  'lrparfldada': Slider(start=0.1, end=10, step=0.1, title='Learning Rate'),
                  }
        uipars = {'uiobjects': uiobjs, }
    else:
        uiobjs = uipars['uiobjects']

    uiobjs['estparfldada'].value = dict['est']
    uiobjs['lrparfldada'].value = dict['lr']
    return uipars


def ensembleChgCB(attrnm, old, new, cb, ensemblegrp):
    ret = uibokeh.getAllUiFlds(ensemblegrp)
    for uifld in ret:
        uifld.visible = False
    ret = uibokeh.getGroup(new, cb.options, ensemblegrp, 'uiobjects')
    for uifld in ret:
        ret[uifld].visible = True


def getUiPars(isclassification, uipars=None):
    models = getUiModelTypes(isclassification)
    uiobjs = {}
    parsgrp = None
    linearkey = 'loggrp' if isclassification else 'lineargrp'
    deftype = classmltypes[0][1] if isclassification else regmltypes[1][1]

    if not uipars:
        uiobjs = {
            'modeltyp': Select(title='Type', options=models),
            linearkey: getLogGrp() if isclassification else getLinearGrp(),
            'ensemblegrp': getEnsembleGrp(),
            'nngrp': getNNGrp(),
            'svmgrp': getSVMGrp()
        }
        modelsgrp = (uiobjs[linearkey], uiobjs['ensemblegrp'],
                     uiobjs['nngrp'], uiobjs['svmgrp'])
        uiobjs['modeltyp'].on_change('value', partial(
            modelChgCB, cb=uiobjs['modeltyp'], modelsgrp=modelsgrp))
        pars = [uiobjs['modeltyp']]
        ensemblepars = [uiobjs['ensemblegrp']['uiobjects']['ensembletyp']]
        if hasXGBoost():
            xgdtpars = [uiobjs['ensemblegrp']['uiobjects']['estparfldxgdt'],
                        uiobjs['ensemblegrp']['uiobjects']['depparfldxgdt'],
                        uiobjs['ensemblegrp']['uiobjects']['lrparfldxgdt'],
                        ]
            xgrfpars = [uiobjs['ensemblegrp']['uiobjects']['estparfldxgrf'],
                        uiobjs['ensemblegrp']['uiobjects']['depparfldxgrf'],
                        uiobjs['ensemblegrp']['uiobjects']['lrparfldxgrf'],
                        ]
            ensemblepars.extend(xgdtpars)
            ensemblepars.extend(xgrfpars)
        ensemblepars.extend([
            uiobjs['ensemblegrp']['uiobjects']['estparfldrf'],
            uiobjs['ensemblegrp']['uiobjects']['depparfldrf'],
            uiobjs['ensemblegrp']['uiobjects']['estparfldgb'],
            uiobjs['ensemblegrp']['uiobjects']['depparfldgb'],
            uiobjs['ensemblegrp']['uiobjects']['lrparfldgb'],
            uiobjs['ensemblegrp']['uiobjects']['estparfldada'],
            uiobjs['ensemblegrp']['uiobjects']['lrparfldada'],
        ])
        nnpars = [uiobjs['nngrp']['uiobjects']['nntyp'],
                  uiobjs['nngrp']['uiobjects']['itrparfld'],
                  uiobjs['nngrp']['uiobjects']['lrparfld'],
                  uiobjs['nngrp']['uiobjects']['lay1parfld'],
                  uiobjs['nngrp']['uiobjects']['lay2parfld'],
                  uiobjs['nngrp']['uiobjects']['lay3parfld'],
                  uiobjs['nngrp']['uiobjects']['lay4parfld'],
                  uiobjs['nngrp']['uiobjects']['lay5parfld'],
                  uiobjs['nngrp']['uiobjects']['buttonparfld'],
                  ]
        svmpars = [uiobjs['svmgrp']['uiobjects']['svmtyp'],
                   uiobjs['svmgrp']['uiobjects']['kernel'],
                   uiobjs['svmgrp']['uiobjects']['degree']
                   ]

        if isclassification:
            pars.extend([uiobjs[linearkey]['uiobjects']['logtyp'],
                         uiobjs[linearkey]['uiobjects']['solvertyp']])
        else:
            pars.extend([uiobjs[linearkey]['uiobjects']['lineartyp']])
        pars.extend(ensemblepars)
        pars.extend(nnpars)
        pars.extend(svmpars)
        parsgrp = column(*pars)
        uipars = {'grp': parsgrp, 'uiobjects': uiobjs}
    else:
        uiobjs = uipars['uiobjects']

    if isclassification:
        uiobjs['modeltyp'].value = models[0]
        uiobjs[linearkey] = getLogGrp(uiobjs[linearkey])
    else:
        uiobjs['modeltyp'].value = models[1]
        uiobjs[linearkey] = getLinearGrp(uiobjs[linearkey])
    uiobjs['ensemblegrp'] = getEnsembleGrp(uiobjs['ensemblegrp'])
    uiobjs['nngrp'] = getNNGrp(uiobjs['nngrp'])
    uiobjs['svmgrp'] = getSVMGrp(uiobjs['svmgrp'])

    modelsgrp = (uiobjs[linearkey], uiobjs['ensemblegrp'],
                 uiobjs['nngrp'], uiobjs['svmgrp'])
    modelChgCB('value', deftype, deftype, uiobjs['modeltyp'], modelsgrp)
    return uipars


def getUiParams(sklearnpars):
    sklearngrp = sklearnpars['uiobjects']
    modeltype = sklearngrp['modeltyp']
    if modeltype.value == 'Linear':
        parmobj = sklearngrp['lineargrp']['uiobjects']
        return getLinearPars(modelname=parmobj['lineartyp'].value)
    if modeltype.value == 'Logistic':
        parmobj = sklearngrp['loggrp']['uiobjects']
        return getLogPars(modelname=parmobj['logtyp'].value,
                          solver=parmobj['solvertyp'].value)
    if modeltype.value == 'Ensemble':
        parmobj = sklearngrp['ensemblegrp']['uiobjects']
        if parmobj['ensembletyp'].value == 'XGBoost: (Decision Tree)':
            return getEnsembleParsXGDT(modelname=parmobj['ensembletyp'].value,
                                       maxdep=parmobj['depparfldxgdt'].value,
                                       est=parmobj['estparfldxgdt'].value,
                                       lr=parmobj['lrparfldxgdt'].value)
        elif parmobj['ensembletyp'].value == 'XGBoost: (Random Forests)':
            return getEnsembleParsXGRF(modelname=parmobj['ensembletyp'].value,
                                       maxdep=parmobj['depparfldxgrf'].value,
                                       est=parmobj['estparfldxgrf'].value,
                                       lr=parmobj['lrparfldxgrf'].value)
        elif parmobj['ensembletyp'].value == 'Random Forests':
            return getEnsembleParsRF(modelname=parmobj['ensembletyp'].value,
                                     maxdep=parmobj['depparfldrf'].value,
                                     est=parmobj['estparfldrf'].value)
        elif parmobj['ensembletyp'].value == 'Gradient Boosting':
            return getEnsembleParsGB(modelname=parmobj['ensembletyp'].value,
                                     maxdep=parmobj['depparfldgb'].value,
                                     est=parmobj['estparfldgb'].value,
                                     lr=parmobj['lrparfldgb'].value)
        elif parmobj['ensembletyp'].value == 'Adaboost':
            return getEnsembleParsAda(modelname=parmobj['ensembletyp'].value,
                                      est=parmobj['estparfldada'].value,
                                      lr=parmobj['lrparfldada'].value)
    elif modeltype.value == 'Neural Network':
        parmobj = sklearngrp['nngrp']['uiobjects']
        return getNNPars(modelname=parmobj['nntyp'].value,
                         maxitr=parmobj['itrparfld'].value,
                         lr=parmobj['lrparfld'].value/1000,
                         lay1=parmobj['lay1parfld'].value,
                         lay2=parmobj['lay2parfld'].value,
                         lay3=parmobj['lay3parfld'].value,
                         lay4=parmobj['lay4parfld'].value,
                         lay5=parmobj['lay5parfld'].value,
                         nb=sklearngrp['nngrp']['nb'].value)
    elif modeltype.value == 'SVM':
        parmobj = sklearngrp['svmgrp']['uiobjects']
        return getSVMPars(modelname=parmobj['svmtyp'].value,
                          kernel=parmobj['kernel'].value,
                          degree=parmobj['degree'].value)
    return None
